from xml.dom.minidom import Attr

from attr import Attribute
from DMCpy import DataSet
from DMCpy import DataFile
import os.path
import matplotlib.pyplot as plt
import numpy as np

def test_init():
    ds = DataSet.DataSet()

    assert(len(ds)==0)

    df = DataFile.loadDataFile(os.path.join('data','dmc2021n{:06d}.hdf'.format(494)))

    ds2 = DataSet.DataSet([df])
    assert(len(ds2)==1)

    ds3 = DataSet.DataSet(dataFiles=df)

    print(ds3.__dict__)
    print(ds2.__dict__)

    assert(ds2==ds3)


def test_load():

    fileNumbers = [494,494,494]
    dataFiles = [os.path.join('data','dmc2021n{:06d}.hdf'.format(no)) for no in fileNumbers]

    ds = DataSet.DataSet(dataFiles)
    
    assert(len(ds) == len(dataFiles))
    assert(ds[0].fileName == os.path.split(dataFiles[0])[-1])


    # load single file and check that it is equal to the corresponding file in ds
    ds2 = DataSet.DataSet(dataFiles[-1])
    assert(len(ds2) == 1)
    assert(ds2[0].fileName == os.path.split(dataFiles[-1])[-1])
    assert(ds2[0] == ds[-1])

    # Find length before appending
    length = len(ds)
    ds.append(dataFiles)

    # Length is supposed to be both
    assert(len(ds)==length+len(dataFiles))

    # Also works for adding data files directly (both in list and as object)
    df = ds2[0]
    ds.append([df])
    assert(len(ds)==length+len(dataFiles)+1)
    ds.append(df)
    assert(len(ds)==length+len(dataFiles)+2)

    # Deletion
    del ds[-1]
    assert(len(ds)==length+len(dataFiles)+1)

def test_plot():

    fileNumbers = [565]
    dataFiles = [os.path.join('data','dmc2021n{:06d}.hdf'.format(no)) for no in fileNumbers]


    ds = DataSet.DataSet(dataFiles)
    ds.monitor[0] = np.array([1.0])
    
    Ax, bins, intensity, error, monitor = ds.plotTwoTheta()

    Ax,*_ = ds.plotTwoTheta(correctedTwoTheta=False)

    # Calculate bins, intensity, error without plotting

    bins2, intensity2, error2, monitor2 = ds.sumDetector()

    print(np.sum(intensity-intensity2))

    assert(np.all(np.isclose(bins,bins2)))
    assert(np.all(np.isclose(intensity,intensity2,equal_nan=True)))
    assert(np.all(np.isclose(error,error2,equal_nan=True)))
    assert(np.all(np.isclose(monitor,monitor2)))
    

def test_2d():
    fileNumbers = [494,494]
    dataFiles = [os.path.join('data','dmc2021n{:06d}.hdf'.format(no)) for no in fileNumbers]

    ds = DataSet.DataSet(dataFiles=dataFiles)

    files = len(fileNumbers)
    assert(ds.counts.shape == (files,1,128,1152))

    ax1 = ds.plotTwoTheta(correctedTwoTheta=False)
    ax2 = ds.plotTwoTheta(correctedTwoTheta=True)


def test_kwargs():
    
    fileNumbers = [494,494]
    dataFiles = [os.path.join('data','dmc2021n{:06d}.hdf'.format(no)) for no in fileNumbers]

    ds = DataSet.DataSet(dataFiles=dataFiles)

    try:
        _ = ds.sumDetector(corrected=False)
        assert(False)
    except AttributeError as e:
        assert(e.args[0] == 'Key-word argument "corrected" not understood. Did you mean "correctedTwoTheta"?')

    ds = DataSet.DataSet(dataFiles=dataFiles)

    try:
        _ = ds.plotTwoTheta(corrected=False,fmt='.-')
        assert(False)
    except AttributeError as e:
        assert(e.args[0] == 'Key-word argument "corrected" not understood. Did you mean "correctedTwoTheta"?')



def test_export_PSI_format(folder='data'):  
    
    if folder is None:
        folder = os.getcwd()
        
    fileNumbers = [565]
    
    dataFiles = [os.path.join(folder,'dmc2021n{:06d}.hdf'.format(no)) for no in fileNumbers]
    print(dataFiles)
    ds = DataSet.DataSet(dataFiles)
    
    for df in ds:
        if np.any(np.isnan(df.monitor)) or np.any(np.isclose(df.monitor,0.0)):
            df.monitor = np.ones_like(df.monitor)
    
    ds.export_PSI_format(outFile="testfile")
    
    assert(os.path.exists("testfile.dat") == True and os.stat("testfile.dat").st_size != 0)
    
    os.remove("testfile.dat")

#test_export_PSI_format()    


def test_export_xye_format(folder='data'):
    
    if folder is None:
        folder = os.getcwd()
        
    fileNumbers = [565]
    
    dataFiles = [os.path.join(folder,'dmc2021n{:06d}.hdf'.format(no)) for no in fileNumbers]
    
    ds = DataSet.DataSet(dataFiles)
    
    for df in ds:
        if np.any(np.isnan(df.monitor)) or np.any(np.isclose(df.monitor,0.0)):
            df.monitor = np.ones_like(df.monitor)
    
    ds.export_xye_format(outFile="testfile")
    
    assert(os.path.exists("testfile.xye") == True and os.stat("testfile.xye").st_size != 0)
    
    os.remove("testfile.xye")

#test_export_xye_format() 



def test_add():
    
    DataSet.add(565,566,outFile='test_add',folder='data')
    
    assert(os.path.exists("test_add.dat") == True and os.stat("test_add.dat").st_size != 0)
    assert(os.path.exists("test_add.xye") == True and os.stat("test_add.xye").st_size != 0)
    os.remove("test_add.dat")
    os.remove("test_add.xye")

# test_add()    

def test_export():
    
    DataSet.export(565,outFile='test_export',folder='data')
    
    assert(os.path.exists("test_export.dat") == True and os.stat("test_export.dat").st_size != 0)
    assert(os.path.exists("test_export.xye") == True and os.stat("test_export.xye").st_size != 0)
    os.remove("test_export.dat")
    os.remove("test_export.xye")

# test_export()


def test_export_from(folder = 'data'):
    """
    Runs the last two files in the data folder. These files must be powder scans for test to work    
    """

    if folder is None:
        folder = os.getcwd()
        
    hdf_files = [f for f in os.listdir(folder) if f.endswith('.hdf')]
    last_hdf = hdf_files[-1]
    numberOfFiles = int(last_hdf.strip('.hdf').split('n')[-1])
    file1 = f"DMC_{numberOfFiles-2}"
    file2 = f"DMC_{numberOfFiles-1}"
 
    DataSet.export_from(numberOfFiles-2,sampleName=False,temperature=False,folder='data')
    
    assert(os.path.exists(file1+'.dat') == True and os.stat(file1+'.dat').st_size != 0)
    assert(os.path.exists(file2+'.dat') == True and os.stat(file2+'.dat').st_size != 0)   
    assert(os.path.exists(file1+'.xye') == True and os.stat(file1+'.xye').st_size != 0)
    assert(os.path.exists(file2+'.xye') == True and os.stat(file2+'.xye').st_size != 0)       
    os.remove(file1+'.dat')
    os.remove(file2+'.dat')
    os.remove(file1+'.xye')
    os.remove(file2+'.xye')


def test_export_from_to():
    
    DataSet.export_from_to(565,566,sampleName=False,temperature=False,folder='data')
    
    assert(os.path.exists("DMC_565.dat") == True and os.stat("DMC_565.dat").st_size != 0)
    assert(os.path.exists("DMC_565.xye") == True and os.stat("DMC_565.xye").st_size != 0)
    assert(os.path.exists("DMC_566.dat") == True and os.stat("DMC_566.dat").st_size != 0)
    assert(os.path.exists("DMC_566.xye") == True and os.stat("DMC_566.xye").st_size != 0)
    os.remove("DMC_565.dat")
    os.remove("DMC_565.xye")
    os.remove("DMC_566.dat")
    os.remove("DMC_566.xye")

# test_export_from_to()


def test_export_list():
    
    DataSet.export([565],outFile='test_export_list',folder='data')
    
    assert(os.path.exists("test_export_list.dat") == True and os.stat("test_export_list.dat").st_size != 0)
    assert(os.path.exists("test_export_list.xye") == True and os.stat("test_export_list.xye").st_size != 0)
    os.remove("test_export_list.dat")
    os.remove("test_export_list.xye")
    
# test_export_list()

def test_subtract_PSI():
    
    DataSet.subtract_PSI('DMC_565','DMC_566',outFile='test_subtract_PSI',folder='data')
    
    assert(os.path.exists("test_subtract_PSI.dat") == True and os.stat("test_subtract_PSI.dat").st_size != 0)
    os.remove("test_subtract_PSI.dat")

def test_subtract_xye():
    
    DataSet.subtract_xye('DMC_565','DMC_566',outFile='test_subtract_xye',folder='data')
    
    assert(os.path.exists("test_subtract_xye.xye") == True and os.stat("test_subtract_xye.xye").st_size != 0)
    os.remove("test_subtract_xye.xye")


def test_subtract():
    
    DataSet.subtract('DMC_565','DMC_566',outFile='test_subtract',folder='data')

    assert(os.path.exists("test_subtract.dat") == True and os.stat("test_subtract.dat").st_size != 0)
    assert(os.path.exists("test_subtract.xye") == True and os.stat("test_subtract.xye").st_size != 0)
    os.remove("test_subtract.dat")
    os.remove("test_subtract.xye")

def test_updateDataFileParameters():


    fileNumbers = [565,565,566]
    dataFiles = [os.path.join('data','dmc2021n{:06d}.hdf'.format(no)) for no in fileNumbers]

    ds = DataSet.DataSet(dataFiles)
    oldWavelength = [df.wavelength for df in ds]

    newWavelength = 10
    ds.updateDataFiles('wavelength',newWavelength)
    assert(np.all([np.isclose(newWavelength,df.wavelength) for df in ds]))

    try:
        ds.updateDataFiles('WaveLength',newWavelength) # Doesn't have the attribute
        assert False
    except AttributeError:
        assert True

    try:
        ds.updateDataFiles('wavelength',[10,10]) # Wrong length of parameters
        assert False
    except AttributeError:
        assert True


    newValues = np.random.rand(len(ds))
    ds.updateDataFiles('twoThetaPosition',newValues)
    assert(np.all([np.isclose(nV,df.twoThetaPosition) for nV,df in zip(newValues,ds)]))



import sys
sys.path.append(r'C:\Software\DMCpy\DMCpy')
from Tutorial_Class import Tutorial
import os

def Tester():
    from DMCpy import DataSet,DataFile
    import matplotlib.pyplot as plt
    import numpy as np
    import os
    
    folder = 'data'
    file = r'dmc2021n009003.hdf'
    
    df = DataFile.loadDataFile(os.path.join(folder,file))
    
    # Use above data file in data set. Must be inserted as a list
    ds = DataSet.DataSet([df])
    
    ds.autoAlignScatteringPlane(scatteringNormal=np.array([1,-1,0],dtype=float))

    Viewer = ds.Viewer3D(0.03,0.03,0.03)
    
    # Set the color bar limits to 0 and 60
    Viewer.set_clim(0,20)
    
    
    # Find the number of steps and set viewer to middel value
    # This can also be done interactively in the viewer by pressing up or down,
    # or by scrolling the mouse wheel or clicking the sliding bar.
    zSteps = Viewer.Z.shape[-1]
    Viewer.setPlane(int(zSteps/2)-1)
    
    fig = Viewer.ax.get_figure()
    fig.savefig(r'docs/Tutorials/View3D/CenterMiddel_cut.png',format='png',dpi=300)
    
    # First we do a cut over the (440) reflection by the cut1D function. 
    # cut1D takes start and end point as lists.
    positionVector,I,ax = ds.plotCut1D([3.8,3.8,0],[4.2,4.2,0],width=0.2,widthZ=0.2)
    fig = ax.get_figure()
    fig.savefig(r'docs/Tutorials/View3D/Cut1.png',format='png',dpi=300)
    
    # Cut over (004) to (008)
    positionVector,I,ax = ds.plotCut1D([0.3,0.2,3],[0.4,0.3,8.1],width=0.5,widthZ=0.2)
    fig = ax.get_figure()
    fig.savefig(r'docs/Tutorials/View3D/Cut2.png',format='png',dpi=300)

    # Cut over (4-x,4-x,x)
    positionVector,I,ax = ds.plotCut1D([4.2,4.2,-0.2],[-0.2,-0.2,4.2],width=0.5,widthZ=0.3)
    fig = ax.get_figure()
    fig.savefig(r'docs/Tutorials/View3D/Cut3.png',format='png',dpi=300)
   
    
    
    
title = 'Cut3D'

introText = 'After inspecting the scattering plane, we want to perform cuts along certain directions.'\
+' In this tutorial, we demonstrate the cut1D function. Cuts can be made given by hkl or Qx, Qy, Qz.'\
+' The width of the cut can be adjusted by the keywords width and widthZ.'


outroText = 'The above code takes the data from the A3 scan file dmc2021n000590, align and plot the scattering plane.'\
+'Then three cuts along different directions are performed.'\
+'\n\nFirst data overview with Qz slightly positive and Qx and Qy in the plane\n'\
+'\n.. figure:: CenterMiddel_cut.png \n  :width: 50%\n  :align: center\n\n '\
+'\n\nFirst cut\n'\
+'\n.. figure:: Cut1.png \n  :width: 50%\n  :align: center\n\n '\
+'\n\nSecond cut\n'\
+'\n.. figure:: Cut2.png \n  :width: 50%\n  :align: center\n\n '\
+'\n\nThrid cut\n'\
+'\n.. figure:: Cut3.png \n  :width: 50%\n  :align: center\n\n '

introText = title+'\n'+'^'*len(title)+'\n'+introText


    
Example = Tutorial('Cut3D',introText,outroText,Tester,fileLocation = os.path.join(os.getcwd(),r'docs/Tutorials/View3D'))

def test_Cut3D():
    Example.test()

if __name__ == '__main__':
    Example.generateTutorial()


import h5py as hdf
import datetime
from h5py._hl import attrs
import numpy as np
import pickle as pickle
import matplotlib.pyplot as plt
import pandas as pd
import DMCpy
import os.path
from DMCpy import InteractiveViewer
from collections import defaultdict
import warnings

import copy
from DMCpy import _tools
from DMCpy import Sample

scanTypes = ['Old Data','Powder','A3']

def decode(item):
    """Test and decode item to utf8"""
    if hasattr(item,'__len__'):
        if len(item)>0:
            if hasattr(item[0],'decode'):
                item = item[0].decode('utf8')
            
                    
    return item


@_tools.KwargChecker()
def maskFunction(phi,maxAngle=10.0):
    """Mask all phi angles outside plus/minus maxAngle

    Args:

        - phi (array): Numpy array of phi angles to be masked

    Kwargs:

        - maxAngle (float): Mask points greater than this or less than -maxAngle in degrees (default 10)
    """
    return np.abs(phi)>maxAngle

@_tools.KwargChecker()
def findCalibration(fileName):
    """Find detector calibration for specified file

    Args:

        - fileName (str): Name of file for which calibration is needed

    Returns:

        - calibration (array): Detector efficiency

        - calibrationName (str): Name of calibration file used

    Raises:

        - fileNotFoundError

    """

    # Extract only actual file name if path is provided    
    fileName = os.path.split(fileName)[-1]

    # Split name in 'dmcyyyynxxxxxx.hdf'
    year,fileNo = [int(x) for x in fileName[3:].replace('.hdf','').split('n')]

    calibrationDict = DMCpy.calibrationDict

    # Calibration files do not cover the wanted year
    if not year in calibrationDict.keys():
        raise FileNotFoundError('Calibration files for year {} (extracted from file name "{}") is'.format(year,fileName)+\
            ' not covered in calibration tables. Please update to newest version by invoking "pip install --upgrade DMCpy"')

    yearCalib = calibrationDict[year]
    
    limits = yearCalib['limits']
    
    # Calibration name is index of the last limit below file number
    idx = np.sum([fileNo>limits])-1
    
    idx = np.max([idx,0]) # ensure that idx is not negative
    
    # Calibration is saved in yearCalib with name of calibration file
    calibrationName = yearCalib['names'][idx]
    calibration = yearCalib[calibrationName]
    return calibration,calibrationName


## Dictionary for holding hdf position of attributes. HDFTranslation['a3'] gives hdf position of 'a3'
HDFTranslation = {'sample':'/entry/sample',
                  'sampleName':'/entry/sample/name',
                  'monitor':'entry/monitor/monitor',
                  'unitCell':'/entry/sample/unit_cell',
                  'counts':'entry/DMC/detector/data',
                  'summedCounts': 'entry/DMC/detector/summed_counts',
                  'monochromatorCurvature':'entry/DMC/monochromator/curvature',
                  'monochromatorVerticalCurvature':'entry/DMC/monochromator/curvature_vertical',
                  'monochromatorGoniometerLower':'entry/DMC/monochromator/goniometer_lower',
                  'monochromatorGoniometerUpper':'entry/DMC/monochromator/goniometer_upper',
                  'monochromatorRotationAngle':'entry/DMC/monochromator/rotation_angle',
                  'monochromatorTakeoffAngle':'entry/DMC/monochromator/takeoff_angle',
                  'monochromatorTranslationLower':'entry/DMC/monochromator/translation_lower',
                  'monochromatorTranslationUpper':'entry/DMC/monochromator/translation_upper',
                  

                  'wavelength':'entry/DMC/monochromator/wavelength',
                  'twoThetaPosition':'entry/DMC/detector/detector_position',
                  'mode':'entry/monitor/mode',
                  'preset':'entry/monitor/preset',
                  'startTime':'entry/start_time',
                  'time':'entry/monitor/time',
                  'endTime':'entry/end_time',
                  'comment':'entry/comment',
                  'proposal':'entry/proposal_id',
                  'proposalTitle':'entry/proposal_title',
                  'localContact':'entry/local_contact/name',
                  'proposalUser':'entry/proposal_user/name',
                  'proposalEmail':'entry/proposal_user/email',
                  'user':'entry/user/name',
                  'email':'entry/user/email',
                  'address':'entry/user/address',
                  'affiliation':'entry/user/affiliation',
                  'A3':'entry/sample/rotation_angle',
                  'temperature':'entry/sample/temperature',
                  'magneticField':'entry/sample/magnetic_field',
                  'electricField':'entry/sample/electric_field',
                  'scanCommand':'entry/scancommand',
                  'title':'entry/title',
                  'absoluteTime':'entry/control/absolute_time',
                  'protonBeam':'entry/proton_beam/data'
}

## Dictionary for holding standard values 

HDFTranslationDefault = {'twoThetaPosition':np.array([0.0]),
                         'comment': 'No Comments',
                         'endTime': '20yy-mm-dd hh:mm:ss',
                         'proposalTitle': 'Unknown Title',
                         'localContact': 'Unknown Local Contact',
                         'proposalUser': 'Unknown User',
                         'proposalEmail': 'Unknown Email',
                         'address': 'Unknown Address',
                         'affiliation': 'Unknown Affiliation',
                         'scanCommand': 'Unknown scanCommand',

                         'temperature': np.array([0.0]),
                         'magneticField': np.array([0.0]),
                         'electricField': np.array([0.0]),

                         'absoluteTime': np.array([0.0]),
                         'protonBeam': np.array([0.0]),
                         
                         

}

## Default dictionary to perform on loaded data, i.e. take the zeroth element, swap axes, etc

HDFTranslationFunctions = defaultdict(lambda : [])
HDFTranslationFunctions['sampleName'] = [['__getitem__',[0]],['decode',['utf8']]]
HDFTranslationFunctions['mode'] = [['__getitem__',[0]],['decode',['utf8']]]
HDFTranslationFunctions['startTime'] = [['__getitem__',[0]],['decode',['utf8']]]
HDFTranslationFunctions['wavelength'] = [['mean',[]]]
HDFTranslationFunctions['twoThetaPosition'] = [['__getitem__',[0]]]
HDFTranslationFunctions['endTime'] = [['__getitem__',[0]]]
HDFTranslationFunctions['experimentalIdentifier'] = [['__getitem__',[0]]]
HDFTranslationFunctions['comment'] = [['__getitem__',[0]],['decode',['utf8']]]
HDFTranslationFunctions['proposal'] = [['__getitem__',[0]]]
HDFTranslationFunctions['proposalTitle'] = [['__getitem__',[0]],['decode',['utf8']]]
HDFTranslationFunctions['localContact'] = [['__getitem__',[0]],['decode',['utf8']]]
HDFTranslationFunctions['proposalUser'] = [['__getitem__',[0]],['decode',['utf8']]]
HDFTranslationFunctions['proposalEmail'] = [['__getitem__',[0]],['decode',['utf8']]]
HDFTranslationFunctions['user'] = [['__getitem__',[0]],['decode',['utf8']]]
HDFTranslationFunctions['email'] = [['__getitem__',[0]],['decode',['utf8']]]
HDFTranslationFunctions['address'] = [['__getitem__',[0]],['decode',['utf8']]]
HDFTranslationFunctions['affiliation'] = [['__getitem__',[0]],['decode',['utf8']]]
HDFTranslationFunctions['scanCommand'] = [['__getitem__',[0]],['decode',['utf8']]]
HDFTranslationFunctions['title'] = [['__getitem__',[0]],['decode',['utf8']]]



HDFInstrumentTranslation = {
}

HDFInstrumentTranslationFunctions = defaultdict(lambda : [])
# HDFInstrumentTranslationFunctions['counts'] = [['swapaxes',[1,2]]]
HDFInstrumentTranslationFunctions['twoThetaPosition'] = [['mean',]]
HDFInstrumentTranslationFunctions['wavelength'] = [['mean',]]

extraAttributes = ['name','fileLocation']

possibleAttributes = list(HDFTranslation.keys())+list(HDFInstrumentTranslation.keys())+extraAttributes
possibleAttributes.sort(key=lambda v: v.lower())

HDFTypes = defaultdict(lambda: lambda x: np.array([np.string_(x)]))
HDFTypes['monitor'] = np.array
HDFTypes['monochromatorCurvature'] = np.array
HDFTypes['monochromatorVerticalCurvature'] = np.array
HDFTypes['monochromatorGoniometerLower'] = np.array
HDFTypes['monochromatorGoniometerUpper'] = np.array
HDFTypes['monochromatorRotationAngle'] = np.array
HDFTypes['monochromatorTakeoffAngle'] = np.array
HDFTypes['monochromatorTranslationLower'] = np.array
HDFTypes['monochromatorTranslationUpper'] = np.array
HDFTypes['wavelength'] = np.array
HDFTypes['twoThetaPosition'] = np.array
# HDFTypes['mode'] = lambda x: np.array([np.string_(x)])
HDFTypes['preset'] = np.array
# HDFTypes['startTime'] = np.string_
HDFTypes['time'] = np.array
# HDFTypes['endTime'] = np.string_
# HDFTypes['comment'] = np.string_
HDFTypes['absoluteTime'] = np.array
HDFTypes['protonBeam'] = np.array


HDFUnits = {
    'monitor':'counts',
    'monochromatorCurvature':'degree',
    'monochromatorVerticalCurvature':'degree',
    'monochromatorGoniometerLower':'degree',
    'monochromatorGoniometerUpper':'degree',
    'monochromatorRotationAngle':'degree',
    'monochromatorTakeoffAngle':'degree',
    'monochromatorTranslationLower':'mm',
    'monochromatorTranslationUpper':'mm',
    'twoThetaPosition':'degree',
    'monitor':'counts',
    'wavelength':'A'
}



def getNX_class(x,y,attribute):
    try:
        variableType = y.attrs['NX_class']
    except:
        variableType = ''
    if variableType==attribute:
        return x

def getInstrument(file):
    location = file.visititems(lambda x,y: getNX_class(x,y,b'NXinstrument'))
    return file.get(location)


@_tools.KwargChecker(include=['radius','twoTheta','verticalPosition','twoThetaPosition']+list(HDFTranslation.keys()))
def loadDataFile(fileLocation=None,fileType='Unknown',**kwargs):
    """Load DMC data file, either powder or single crystal data.
    
    
    """

    if fileLocation is None:
        return DataFile()

    if isinstance(fileLocation,(DataFile)):
        if fileLocation.fileType.lower() == 'powder':
            return PowderDataFile(fileLocation)
        elif fileLocation.fileType.lower() == 'singlecrystal':
            return SingleCrystalDataFile(fileLocation)
        else:
            return DataFile(fileLocation)
    elif not os.path.exists(fileLocation): # load file from disk
        raise FileNotFoundError('Provided file path "{}" not found.'.format(fileLocation))

    A3 = shallowRead([fileLocation],['A3'])[0]['A3']
    
    T = 'Unknown' # type of datafile

    if A3 is None: # there is no A3 values at all
        T = 'powder'
        
    elif len(A3) == 1:
        T = 'powder'
    else:
        T = 'singlecrystal'

    ## here be genius function to determine type of data
    
    if fileType.lower() == 'powder' or T == 'powder':
        df = PowderDataFile(fileLocation)
    elif fileType.lower() == 'singlecrystal' or T == 'singlecrystal':
        df = SingleCrystalDataFile(fileLocation)
    else:
        df = DataFile(fileLocation)

    
    repeats = df.counts.shape[1]
    # Insert standard values if not present in kwargs
    if not 'radius' in kwargs:
        kwargs['radius'] = 0.8

    if not 'verticalPosition' in kwargs:
        kwargs['verticalPosition'] = np.linspace(-0.1,0.1,repeats,endpoint=True)

    # Overwrite parameters provided in the kwargs
    for key,item in kwargs.items():
        setattr(df,key,item)
        
    if 'twoThetaPosition' in kwargs:
        if not 'twoTheta' in kwargs:
            df.twoTheta = np.linspace(0,132,9*128)+df.twoThetaPosition
        else:
            df.twoTheta = kwargs['twoTheta']
    elif 'twoTheta' in kwargs:
        df.twoTheta = kwargs['twoTheta']
    
    df.initializeQ()
    df.loadNormalization()

    return df



class DataFile(object):
    @_tools.KwargChecker()
    def __init__(self, file=None):
        self.fileType = 'DataFile'
        self._twoThetaOffset = 0.0

        if not file is None: 
            if isinstance(file,DataFile): # Copy everything from provided file
                # Copy all file settings
                self.updateProperty(file.__dict__)

            elif os.path.exists(file): # load file from disk
                self.loadFile(file)


            else:
                raise FileNotFoundError('Provided file path "{}" not found.'.format(file))

    
    @_tools.KwargChecker()
    def loadFile(self,filePath):
        if not os.path.exists(filePath):
            raise FileNotFoundError('Provided file path "{}" not found.'.format(filePath))

        self.folder, self.fileName = os.path.split(filePath)

        # setup standard parameters

        self._wavelength = 0.0


        # Open file in reading mode
        with hdf.File(filePath,mode='r') as f:

            self.sample = Sample.Sample(sample=f.get(HDFTranslation['sample']))

            # load standard things using the shallow read
            instr = getInstrument(f)
            for parameter in HDFTranslation.keys():
                if parameter in ['unitCell','sample','unitCell']:
                    continue
                
                if parameter in HDFTranslation:
                    value = np.array(f.get(HDFTranslation[parameter]))
                    TrF= HDFTranslationFunctions
                elif parameter in HDFInstrumentTranslation:
                    value = np.array(instr.get(HDFInstrumentTranslation[parameter]))
                    TrF= HDFInstrumentTranslationFunctions

                if value.shape == () or value is None:
                    value = HDFTranslationDefault[parameter]

                else:
                    for func,args in TrF[parameter]:
                        value = getattr(value,func)(*args)
                
                setattr(self,parameter,value)
                
        self.counts.shape = (1,*self.counts.shape) # Standard shape

    def initializeQ(self):
        if len(self.twoTheta.shape) == 2:
            self.twoTheta, z = np.meshgrid(self.twoTheta[0].flatten(),self.verticalPosition,indexing='xy')
        else:
            self.twoTheta, z = np.meshgrid(self.twoTheta.flatten(),self.verticalPosition,indexing='xy')
            
        self.pixelPosition = np.array([self.radius*np.cos(np.deg2rad(self.twoTheta)),
                                    -self.radius*np.sin(np.deg2rad(self.twoTheta)),
                                    z]).reshape(3,*self.counts.shape[1:])
        
        
        #self.Monitor = self.monitor
        
        if np.any(np.isclose(self.monitor,0)): # error mode from commissioning
            self.monitor = np.ones(self.counts.shape[0])
        
        self.alpha = np.rad2deg(np.arctan2(self.pixelPosition[2],self.radius))
        # Above line makes an implicit call to the self.calculateQ method!
        
        self.calculateQ()
        self.generateMask(maskingFunction=None)


    
    def loadNormalization(self):
        # Load calibration
        try:
            self.normalization, self.normalizationFile = findCalibration(self.fileName)
        except ValueError:
            self.normalizationFile = 'None'

        if self.normalizationFile == 'None':
            self.normalization = np.ones_like(self.counts,dtype=float)
        else:
            
            if self.fileType.lower() == "singlecrystal": # A3 scan
                self.normalization = np.repeat(self.normalization,self.counts.shape[0],axis=0)
                self.normalization.shape = self.counts.shape
            else:
                self.normalization.shape = self.counts.shape

    def __len__(self):
        if hasattr(self,'counts'):
            return len(self.counts)
        else:
            return 0

    @property
    def A3(self):
        return self.sample.rotation_angle

    @A3.getter
    def A3(self):
        if not hasattr(self.sample,'rotation_angle'):
            self.sample.rotation_angle = np.array([0.0]*len(self.monitor))
        return self.sample.rotation_angle
        

    @A3.setter
    def A3(self,A3):
        if A3 is None:
            self.sample.rotation_angle = np.array([0.0]*len(self.monitor))
        else:
            self.sample.rotation_angle = A3
        if hasattr(self,'ki'):
            self.calculateQ()
    

    @property
    def twoThetaPosition(self):
        return self._detector_position

    @twoThetaPosition.getter
    def twoThetaPosition(self):
        if not hasattr(self,'_detector_position'):
            self._detector_position = np.array([0.0])
        return self._detector_position+self.twoThetaOffset

    @twoThetaPosition.setter
    def twoThetaPosition(self,twoTheta):
        if twoTheta is None:
            self._detector_position = np.array([0.0]*len(self.A3))
        elif np.isnan(twoTheta):
            self._detector_position = np.array([0.0]*len(self.A3))
        else:
            self._detector_position = np.asarray(twoTheta)
        self.twoTheta = np.repeat((np.linspace(0,132,1152) + self._detector_position + self._twoThetaOffset)[np.newaxis],self.counts.shape[1],axis=0)
        if hasattr(self,'_Ki') and hasattr(self,'twoTheta'):
            self.calculateQ()

    

    @property
    def Ki(self):
        return self._Ki

    @Ki.getter
    def Ki(self):
        return self._Ki

    @Ki.setter
    def Ki(self,Ki):
        self._Ki = Ki
        self.wavelength = np.full_like(self.wavelength,2*np.pi/Ki)
        self.calculateQ()

    @property
    def twoThetaOffset(self):
        return self._twoThetaOffset

    @twoThetaOffset.getter
    def twoThetaOffset(self):
        return self._twoThetaOffset

    @twoThetaOffset.setter
    def twoThetaOffset(self,dTheta):
        self._twoThetaOffset = dTheta
        self.twoTheta = np.repeat((np.linspace(0,132,1152) + self._detector_position + self._twoThetaOffset)[np.newaxis],self.counts.shape[1],axis=0)
        self.calculateQ()

    @property
    def wavelength(self):
        return self._wavelength

    @wavelength.getter
    def wavelength(self):
        if not hasattr(self,'_wavelength'):
            self._wavelength = 0.0
        return self._wavelength

    @wavelength.setter
    def wavelength(self,wavelength):
        self._wavelength = wavelength
        self._Ki = 2*np.pi/wavelength
        self.calculateQ()

    def calculateQ(self):
        """Calculate Q and qx,qy,qz using the current A3 values"""
        if not (hasattr(self,'Ki') and hasattr(self,'twoTheta')
                and hasattr(self,'alpha') and hasattr(self,'A3')):
            return 
        self.ki = np.array([0.0,self.Ki,0.0]) # along ki=2pi/lambda with x
        self.ki.shape = (3,1,1)

        self.kf = self.Ki * np.array([-np.sin(np.deg2rad(self.twoTheta))*np.cos(np.deg2rad(self.alpha)),
                                    np.cos(np.deg2rad(self.twoTheta))*np.cos(np.deg2rad(self.alpha)),
                                    np.sin(np.deg2rad(self.alpha))])
        self.q = self.ki-self.kf   
        if self.fileType.lower() == 'singlecrystal': # A3 Scan
            # rotate kf to correct for A3
            zero = np.zeros_like(self.A3)
            ones = np.ones_like(self.A3)
            rotMat = np.array([[np.cos(np.deg2rad(self.A3)),np.sin(np.deg2rad(self.A3)),zero],[-np.sin(np.deg2rad(self.A3)),np.cos(np.deg2rad(self.A3)),zero],[zero,zero,ones]])
            q_temp = self.ki-self.kf

            self.q = np.einsum('jki,k...->ji...',rotMat,q_temp)
            self.Q = np.linalg.norm(self.q,axis=0)
        else:
            self.Q = np.array([np.linalg.norm(self.q,axis=0)])

        self.correctedTwoTheta = 2.0*np.rad2deg(np.arcsin(self.wavelength*self.Q[0]/(4*np.pi)))[np.newaxis].repeat(self.Q.shape[0],axis=0)
        self.phi = np.rad2deg(np.arctan2(self.q[2],np.linalg.norm(self.q[:2],axis=0)))
        

    def generateMask(self,maskingFunction = maskFunction, **pars):
        """Generate mask to applied to data in data file
        
        Kwargs:

            - maskingFunction (function): Function called on self.phi to generate mask (default maskFunction)

        All other arguments are passed to the masking function.

        """

        # check if counts attribute is available

        if not hasattr(self,'counts'):
            raise RuntimeError('DataFile does not contain any counts. Look for self.counts but found nothing.')

        if maskingFunction is None:
            self.mask = np.zeros_like(self.counts,dtype=bool)
        else:
            self.mask = maskingFunction(self.phi,**pars)
        
        

    def updateProperty(self,dictionary):
        """Update self with key and values from provided dictionary. Overwrites any properties already present."""
        if isinstance(dictionary,dict):
            for key,item in dictionary.items():
                if key == 'exclude': continue
                if key == 'kwargs': # copy kwargs directly and continue
                    self.kwargs = item
                    continue
                item = decode(item)
                self.__setattr__(key,copy.deepcopy(item))
        else:
            raise AttributeError('Provided argument is not of type dictionary. Received instance of type {}'.format(type(dictionary)))


    @_tools.KwargChecker(function=plt.errorbar,include=_tools.MPLKwargs)
    def plotDetector(self,ax=None,applyNormalization=True,**kwargs):
        """Plot intensity as function of twoTheta (and vertical position of pixel in 2D)

        Kwargs:

            - ax (axis): Matplotlib axis into which data is plotted (default None - generates new)

            - applyNormalization (bool): If true, take detector efficiency into account (default True)

            - All other key word arguments are passed on to plotting routine

        """

        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()

        
        intensity = self.counts/self.monitor.reshape(-1,1,1)
        if applyNormalization:
            intensity=self.intensity/self.monitor.reshape(-1,1,1)#1.0/self.normalization

        count_err = np.sqrt(self.counts)
        intensity_err = count_err/self.monitor.reshape(-1,1,1)
        if applyNormalization:
            intensity_err*=1.0/self.normalization
 


        # If data is one dimensional
        if self.twoTheta.shape[1] == 1:
            if not 'fmt' in kwargs:
                kwargs['fmt'] = '_'

            ax._err = ax.errorbar(self.twoTheta[np.logical_not(self.mask)],intensity[np.logical_not(self.mask)],intensity_err[np.logical_not(self.mask)],**kwargs)
            ax.set_xlabel(r'$2\theta$ [deg]')
            ax.set_ylabel(r'Counts/mon [arb]')

            def format_coord(ax,xdata,ydata):
                if not hasattr(ax,'xfmt'):
                    ax.mean_x_power = _tools.roundPower(np.mean(np.diff(ax._err.get_children()[0].get_data()[0])))
                    ax.xfmt = r'$2\theta$ = {:3.'+str(ax.mean_x_power)+'f} Deg'
                if not hasattr(ax,'yfmt'):
                    ymin,ymax,ystep = [f(ax._err.get_children()[0].get_data()[1]) for f in [np.min,np.max,len]]
                    
                    ax.mean_y_power = _tools.roundPower((ymax-ymin)/ystep)
                    ax.yfmt = r'Int = {:.'+str(ax.mean_y_power)+'f} cts'

                return ', '.join([ax.xfmt.format(xdata),ax.yfmt.format(ydata)])

            ax.format_coord = lambda format_xdata,format_ydata:format_coord(ax,format_xdata,format_ydata)
        else: # plot a 2D image with twoTheta vs z
            # Set all masked out points to Nan
            intensity[self.mask] = np.nan

            if 'colorbar' in kwargs:
                colorbar = kwargs['colorbar']
                del kwargs['colorbar']
            else:
                colorbar = False
            
            ax._pcolormesh = ax.pcolormesh(self.twoTheta,self.pixelPosition[2],np.sum(intensity,axis=0),shading='auto')

            if colorbar:
                ax._col = fig.colorbar(ax._pcolormesh)
                ax._col.set_label('Intensity [cts/Monitor]')
                

            ax.set_xlabel(r'$2\theta$ [deg]')
            ax.set_ylabel(r'z [m]')

        return ax

    @_tools.KwargChecker()
    def save(self,filePath,compression=6):
        """Save data file in hdf format.
        
        Args:

            - filePath (path): Path into which file is to be saved

        Kwargs:

            - compression (int): Compression level used by gzip

        """
        if os.path.exists(filePath):
            raise AttributeError('File already exists! ({})'.format(filePath))

        
        with hdf.File(filePath,'w') as f:
    
            # Create correct header info
            f.attrs['NeXus_Version'] = np.string_('4.4.0')
            f.attrs['file_name'] = np.string_(filePath)
            
            
            cT = datetime.datetime.now()
            
            f.attrs['file_time'] = np.string_('{}-{}-{} {}:{}:{}'.format(cT.year,cT.month,cT.day,cT.hour,cT.minute,cT.second))
            f.attrs['instrument'] = np.string_('DMC')
            f.attrs['owner'] = np.string_('Lukas Keller <lukas.keller@psi.ch>')

            entry = f.create_group('entry')
            entry.attrs['NX_class'] = np.string_('NXentry')
            entry.attrs['default'] = np.string_('data')

            # Generate file structure
            DMC = entry.create_group('DMC')
            DMC.attrs['NX_class'] = np.string_('NXinstrument')
            
            SINQ = DMC.create_group('SINQ')
            SINQ.attrs['NX_class'] = np.string_('NXsource')
            SINQ.attrs['name'] = np.string_('SINQ')
            SINQ.attrs['type'] = np.string_('Continuous flux spallation source')
            
            detector = DMC.create_group('detector')
            detector.attrs['NX_class'] = np.string_('NXdetector')

            if self.fileType.lower() != 'singlecrystal':
                position = detector.create_dataset('detector_position',data=np.array(self.twoThetaPosition))
            else:
                position = detector.create_dataset('detector_position',data=np.fill(len(self),self.twoThetaPosition))
            
            position.attrs['units'] = np.string_('degree')

            summedCounts = detector.create_dataset('summed_counts',data=self.counts.sum(axis=0))
            summedCounts.attrs['units'] = np.string_('counts')
            
            
            # Generate structure of file
            
            
            
            mono = DMC.create_group('monochromator')
            mono.attrs['NX_class'] = np.string_('NXmonochromator')
            mono.attrs['type'] = np.string_('Pyrolytic Graphite')
            
            wavelength = mono.create_dataset('wavelength',data=np.array([self.wavelength]))
            wavelength.attrs['units'] = 'A'
            
            # data
            data = entry.create_group('data')
            data.attrs['NX_class'] = np.string_('NXdata')
            data.attrs['signal'] = np.string_('data')
            
            Monitor = entry.create_group('monitor')
            Monitor.attrs['NX_class'] = np.string_('NXmonitor')
            
            
            user = entry.create_group('user')
            user.attrs['NX_class'] = np.string_('NXuser')
            
            
            for key,value in HDFTranslation.items():
                if key in ['counts','summedCounts','wavelength','detector_position']: continue
                if 'sample' in value: continue
                selfValue = HDFTypes[key](getattr(self,key))
                
                newEntry = f.create_dataset(value,data=selfValue)
                if key in HDFUnits:
                    newEntry.attrs['units'] = np.string_(HDFUnits[key])

                    
            
            sample = entry.create_group('sample')
            sample.attrs['NX_class'] = np.string_('NXsample')
            
            a3 = sample.create_dataset('rotation_angle',data=self.A3)
            a3.attrs['units'] = 'degree'
            
            self.sample.saveToHdf(sample)
            
            
            
            # data
            data = entry['data']
            data.attrs['NX_class'] = np.string_('NXdata')
            data.attrs['signal'] = np.string_('data')
            
            if self.fileType.lower() != 'singlecrystal':
                Data = data.create_dataset('data',data=self.counts[0],compression=compression)
            else:
                Data = data.create_dataset('data',data=self.counts,compression=compression)
            Data.attrs['units'] = np.string_('A')
            
            # Create link to data in the right place
            data = detector['data'] = Data
            data.attrs['signal'] = np.int32(1)
            data.attrs['target'] = np.string_('/entry/DMC/detector/data')

            
            entry['monitor/monitor'].attrs['units'] = np.string_('counts')


    def __eq__(self,other):
        return len(self.difference(other))==0
    
    def difference(self,other,keys = set(['sample.name','wavelength','counts','A3','twoTheta','fileType','monitor'])):
        """Return the difference between two data files by keys"""
        dif = []
        if not set(self.__dict__.keys()) == set(other.__dict__.keys()): # Check if same generation and type (hdf or nxs)
            return list(set(self.__dict__.keys())-set(other.__dict__.keys()))

        comparisonKeys = keys
        for key in comparisonKeys:
            skey = self
            okey = other
            while '.' in key:
                baseKey,*keys = key.split('.')
                skey = getattr(skey,baseKey)
                okey = getattr(okey,baseKey)
                key = '.'.join(keys)
            if isinstance(skey,np.ndarray):
                try:
                    if not np.all(np.isclose(skey,okey)):
                        if not np.all(np.isnan(skey),np.isnan(okey)):
                            dif.append(key)
                except (TypeError, AttributeError,ValueError):
                    if np.all(skey!=okey):
                        dif.append(key)
            elif not np.all(getattr(skey,key)==getattr(okey,key)):
                dif.append(key)
        return dif


    @property
    def intensity(self):
        return np.divide(self.counts,self.normalization)

    def InteractiveViewer(self,**kwargs):
        if not self.scanType.lower() in ['singlecrystal','powder'] :
            raise AttributeError('Interactive Viewer can only be used for the new data files. Either for powder or for a single crystal A3 scan')
        return InteractiveViewer.InteractiveViewer(self.intensity,self.twoTheta,self.pixelPosition,self.A3,scanParameter = 'A3',scanValueUnit='deg',colorbar=True,**kwargs)








class SingleCrystalDataFile(DataFile):
    def __init__(self,fileType):
        super(SingleCrystalDataFile,self).__init__(fileType)
        self.fileType = 'SingleCrystal'
        self.counts.shape = (-1,128,1152)

class PowderDataFile(DataFile):
    def __init__(self,fileType):
        super(PowderDataFile,self).__init__(fileType)
        self.fileType = 'Powder'
        self.counts.shape = (1,128,1152)



def shallowRead(files,parameters):

    parameters = np.array(parameters)
    values = []
    possibleAttributes.sort(key=lambda v: v.lower())
    possible = []
    for p in parameters:
        possible.append(p in possibleAttributes)
    
    if not np.all(possible):
        if np.sum(np.logical_not(possible))>1:
            raise AttributeError('Parameters {} not found'.format(parameters[np.logical_not(possible)]))
        else:
            raise AttributeError('Parameter {} not found'.format(parameters[np.logical_not(possible)]))
    
    for file in files:
        vals = {}
        vals['file'] = file
        with hdf.File(file,mode='r') as f:
            instr = getInstrument(f)
            for p in parameters:
                if p == 'name':
                    v = os.path.basename(file)
                    vals[p] = v
                    continue
                elif p == 'fileLocation':
                    v = os.path.dirname(file)
                    vals[p] = v
                    continue
                elif p in HDFTranslation:
                    v = np.array(f.get(HDFTranslation[p]))
                    TrF= HDFTranslationFunctions
                elif p in HDFInstrumentTranslation:
                    v = np.array(instr.get(HDFInstrumentTranslation[p]))
                    TrF= HDFInstrumentTranslationFunctions
                else:
                    raise AttributeError('Parameter "{}" not found'.format(p))
                for func,args in TrF[p]:
                    try:
                        v = getattr(v,func)(*args)
                    except (IndexError,AttributeError):
                        warnings.warn('Parameter "{}" not found in file "{}"'.format(p,file))
                        v = None
                        
                vals[p] = v
        values.append(vals)

    return values


import sys
sys.path.append(r'C:\Users\lass_j\Documents\Software\DMCpy')
from Tutorial_Class import Tutorial


def Tester():
    from DMCpy import DataSet,DataFile
    import numpy as np
    
    file = r'C:/Users/lass_j/Documents/DMC_2021/dmc2021n009003.hdf'
    
    df = DataFile.loadDataFile(file)
    
    # Use above data file in data set. Must be inserted as a list
    ds = DataSet.DataSet([df])
    
    ds.autoAlignScatteringPlane(scatteringNormal=np.array([1,-1,0],dtype=float))

    Viewer = ds.Viewer3D(0.03,0.03,0.03)
    
    # Set the color bar limits to 0 and 60
    Viewer.set_clim(0,20)
    
    
    # Find the number of steps and set viewer to middel value
    # This can also be done interactively in the viewer by pressing up or down,
    # or by scrolling the mouse wheel or clicking the sliding bar.
    zSteps = Viewer.Z.shape[-1]
    Viewer.setPlane(int(zSteps/2)-1)
    
    fig = Viewer.ax.get_figure()
    fig.savefig(r'C:/Users/lass_j/Documents/Software/DMCpy/docs/Tutorials/View3D/CenterMiddel.png',format='png',dpi=300)
    
    #  Change programatically to the next plane
    Viewer.setPlane(int(zSteps/2))
    fig2 = Viewer.ax.get_figure()
    fig2.savefig(r'C:/Users/lass_j/Documents/Software/DMCpy/docs/Tutorials/View3D/CenterAboveMiddel.png',format='png',dpi=300)
    
    
    # Instead of only stepping through the data with the Qx and Qy in the plane
    # one can flip the view by clicking 0, 1, or 2 in the interactive view,
    # or do it programmatically by
    
    Viewer.changeAxis(0)
    xSteps = Viewer.X.shape[-1]
    # Notice that the shape of X, Y, and Z changes when the axis is flipped! 
    # The last dimension is alway 'orthogonal' to the view.
    Viewer.setPlane(174)
    
    fig3 = Viewer.ax.get_figure()
    fig3.savefig(r'C:/Users/lass_j/Documents/Software/DMCpy/docs/Tutorials/View3D/CenterQx.png',format='png',dpi=300)
    
    
    
title = 'Viewer 3D'

introText = 'In a single crystal experiment, the first step is to gain an overview of the system. This is most often done '\
+'by performing an A3 scan with the sample in a specific phase. Due to the 2D detector of DMC, such an A3 scan produces '\
+'a 3D set of measured data points. In the frame of reference of the instrument, the majority of the covered volume is '\
+'in the Qx-Qy plane, i.e. with Qz close to zero. A single A3 slices corresponds to a curved line in th Qx-Qy '\
+' together with a symmetrically curved line in Qz. This sheet is then rotated around the origin with each A3 step.'\
+'\n\n'\
+'Interactivity\n'+'-'*len('Interactivity')+'\n'\
+'The Viewer 3D is an interactive plotting function which allows a look through the data in steps along the three major axes. '\
+'In the current build, these are along Qx, Qy, and Qz, and once can change between them by clicking keys on the keyboard '\
+'as tabulated below\n\n'\
+'+-----------------+------------------------------------------------------+ \n'\
+'|    Key press    | Effect                                               | \n'\
+'+-----------------+------------------------------------------------------+ \n'\
+'|        0        | Change plotting such that Qx is constant             | \n'\
+'+-----------------+------------------------------------------------------+ \n'\
+'|        1        | Change plotting such that Qy is constant             | \n'\
+'+-----------------+------------------------------------------------------+ \n'\
+'|        2        | Change plotting such that Qz is constant             | \n'\
+'+-----------------+------------------------------------------------------+ \n'\
+'| \+ or UpArrow   | Increment step along constant axis                   | \n'\
+'+-----------------+------------------------------------------------------+ \n'\
+'|  Page Up        | Increment 10 steps along constant axis               | \n'\
+'+-----------------+------------------------------------------------------+ \n'\
+'|\- or DownArrow  | Decrement step along constant axis                   | \n'\
+'+-----------------+------------------------------------------------------+ \n'\
+'|  Page Down      | Decrement 10 steps along constant axis               | \n'\
+'+-----------------+------------------------------------------------------+ \n'\
+'|        s        | Save current figure                                  | \n'\
+'+-----------------+------------------------------------------------------+ \n'\
+'|       End       | Skip to the end of allowed stop in this projection   | \n'\
+'+-----------------+------------------------------------------------------+ \n'\
+'|       Home      | Skip to the start of allowed stop in this projection | \n'\
+'+-----------------+------------------------------------------------------+ \n'\


outroText = 'The above code takes the data from the A3 scan file dmc2021n000590, changes the A3 range '\
+'to the  correct one and generates the Viewer 3D utilizing a voxel size of 0.03 x 0.03 x 0.03 A:math:`^{-3}`. '\
+'By default, the viewer starts out in projection 2, i.e. with Qz being the axis stepped through. When handling the '\
+'data directly it is more convenient to utilize the keyboard shortcuts but in a scripting interface these are not '\
+'available. Instead one can utilize the *.setPlane* and *.changeAxis* methods. In addition, the color scale has been tweaked '\
+'such that weaker peaks are visible. It is possible to slightly tweak the color scale directly in the Viewer 3D by using the '\
+'sliders to the right of the color bar. Notice: When saving the figure these slides are rendered invisible. The end results are shown below:'\
+'\n\nFirst data overview with Qz slightly positive and Qx and Qy in the plane\n'\
+'\n.. figure:: CenterMiddel.png \n  :width: 50%\n  :align: center\n\n '\
+'\n\nOne step \'higher\' up along Qz in the same scattering plane\n'\
+'\n.. figure:: CenterAboveMiddel.png \n  :width: 50%\n  :align: center\n\n '\
+'\n\nFlipping of the scattering plane axis to 0, i.e. with Qx being constant\n'\
+'\n.. figure:: CenterQx.png \n  :width: 50%\n  :align: center\n\n '


introText = title+'\n'+'^'*len(title)+'\n'+introText


    
Example = Tutorial('Viewer3D',introText,outroText,Tester,fileLocation = r'C:/Users/lass_j/Documents/Software/DMCpy/docs/Tutorials/View3D')

def test_Viewer3D():
    Example.test()

if __name__ == '__main__':
    Example.generateTutorial()


import argparse
import glob
import os
import numpy as np

OpossibleCommands = ['Nope','None','I don\'t know']

list_of_commands = glob.glob(os.path.join('DMCpy','CommandLineScripts','*.py')) # * means all if need specific format then *.csv

def oxfordlist(ls):
    if len(ls) == 1:
        return ls[0]
    if len(ls) == 2:
        return ' and '.join(ls)
    return ', '.join(ls[-1])+'and '+ls[-1] 

## python files to skip
skip = ['DMCpy.py','__init__.py']

list_of_valid_commands = [command for command in list_of_commands if not os.path.split(command)[-1] in skip]
list_of_valid_command_names = [os.path.splitext(os.path.split(command)[-1])[0] for command in list_of_valid_commands]

parser = argparse.ArgumentParser(description="Collection of DMCpy tools for the command line")
parser.add_argument("task", nargs='?', default='help', type=str, help="Type of task to be performed. Possible tasks are: {}. Run without argument to see help menu.".format(oxfordlist(list_of_valid_command_names)))
parser.add_argument('additional', nargs=argparse.REMAINDER)

args = parser.parse_args()

try:
    idx = list_of_valid_command_names.index(args.task)

    arguments = dict(vars(args))
    del arguments['task']

    print('wanting to call {} with arguments {}'.format(list_of_valid_commands[idx],arguments))
    os.system('python {} {}'.format(list_of_valid_commands[idx],args.additional))
except:
    print('Command not understood from DMCpy')
    print(args)




import glob
import os


def getLatestWheel():
    list_of_files = glob.glob('dist/*') # * means all if need specific format then *.csv
    latest_file = max(list_of_files, key=os.path.getctime)
    return latest_file

def extractVersionFromWheel(fileName=None):
    if fileName is None:
        fileName = getLatestWheel()
    latestVersion = fileName.split('-')[-1].replace('.tar.gz','')
    return latestVersion

def getVersionInSetup():
    with open('setup.py') as f:
        lines = f.readlines()
    for l in lines:
        if l.find("    version='")!=-1:
            version = l.split('=')[-1][1:-1]
            break

    return version


latestVersion = extractVersionFromWheel()
setupVersion = getVersionInSetup()
if latestVersion != setupVersion:
    os.system("python setup.py sdist")

wheel = getLatestWheel()
os.system("twine upload {} -r pypi".format(wheel))


import numpy as np
from DMCpy import _tools

def cosd(x):
    return np.cos(np.deg2rad(x))

def sind(x):
    return np.sin(np.deg2rad(x))

def tand(x):
    return np.tan(np.deg2rad(x))

def arccosd(x):
    return np.rad2deg(np.arccos(x))

def arcsind(x):
    return np.rad2deg(np.arcsin(x))

def arctan2d(y,x):
    return np.rad2deg(np.arctan2(y,x))

factorsqrtEK = 0.694692

def uFromAngles(om,sgu,sgl):
    return np.array([cosd(om)*cosd(sgl),-sind(om)*cosd(sgu)+cosd(om)*sind(sgl)*sind(sgu),
                      sind(om)*sind(sgu)+cosd(om)*sind(sgl)*cosd(sgu)]).T
    
    
def calcUBFromAngles(B, om, sgu, sgl):
    N = np.array([[1.0,0,0],[0,cosd(sgu),-sind(sgu)],[0,sind(sgu),cosd(sgu)]])
    M = np.array([[cosd(sgl),0,sind(sgl)],[0,1,0],[-sind(sgl),0,cosd(sgl)]])    
    OM= np.array([[cosd(om),-sind(om),0],[sind(om),cosd(om),0],[0,0,1]])
    
    UB = np.dot(OM,M)
    UB = np.dot(UB,N)
    UB = np.dot(UB,B)

    return UB  

def calculateBMatrix(cell):
    [a1,a2,a3,b1,b2,b3,alpha1,alpha2,alpha3,beta1,beta2,beta3] = cell
    
    B = np.array([[b1,b2*cosd(beta3),b3*cosd(beta2)],
              [0,b2*sind(beta3),-b3*sind(beta2)*cosd(alpha1)],
              [0,0,b3]]) # 2*np.pi/a3
    return B

def matFromTwoVectors(v1,v2):
    a1 = v1/np.linalg.norm(v1)
    a2 = v2/np.linalg.norm(v2)
    a3 = np.cross(a1,a2)
    a3/=np.linalg.norm(a3)
    a2 = np.cross(a1,a3)
    return np.array([a1,a2,a3]).T


def calcTheta(ki,kf,stt):
    return arctan2d(np.abs(ki) - np.abs(kf) * cosd(stt), #rtan
              np.abs(kf) * sind(stt));
    
def calcTasUVectorFromAngles(r):
    A3 = r[3]
    A4 = r[4]
    ss = np.sign(A4) # Scattering sense
    stt = np.abs(A4) # Sample 2theta
    Ei = r[7]
    Ef = r[8]
    ki = np.sqrt(Ei)*factorsqrtEK
    kf = np.sqrt(Ef)*factorsqrtEK
    sgu = r[5]
    sgl = r[6]
    
    theta = calcTheta(ki, kf, stt)
    try:
        om = A3-ss*theta
    except:
        om = A3.reshape(-1,1,1)-ss*theta
    return uFromAngles(om,sgu,sgl)
  
def calcTasUBFromTwoReflections(cell, r1, r2):
    B = calculateBMatrix(cell)
    R1 = np.array(r1[:3])
    R2 = np.array(r2[:3])
    h1 = np.dot(B,R1)
    h2 = np.dot(B,R2)
    
    HT = matFromTwoVectors(h1, h2)
    u1 = calcTasUVectorFromAngles(r1)
    u2 = calcTasUVectorFromAngles(r2)
    
    UT = matFromTwoVectors(u1, u2)
    HTT = HT.T
    
    U = np.dot(UT,HTT)
    UB = np.dot(U,B)
    return UB
    
def tasReflectionToQC(qe,UB):
    qe = qe[:3]
    return np.dot(UB,qe)

def buildTVMatrix(U1V,U2V):
    U1V /=np.linalg.norm(U1V)
    U2V /=np.linalg.norm(U2V)
    T3V = np.cross(U1V.T, U2V.T).T
    
    return np.array([U1V,U2V,T3V])

def buildRMatrix(UB, planeNormal, qe):
    U1V = tasReflectionToQC(qe,UB)
    U1V/=np.linalg.norm(U1V)
    U2V = np.cross(planeNormal.T, U1V.T).T
    if np.linalg.norm(U2V)<0.001:
        raise AttributeError('Calculate length of U2V too small ({})'.format(U2V))
    
    TV = buildTVMatrix(U1V, U2V)
    TVinv = np.linalg.inv(TV)
    return TVinv

def calcTasQAngles(UB,planeNormal,ss,A3Off,qe):
    R = buildRMatrix(UB,planeNormal,qe)
    cossgl = np.sqrt(R[0,0]*R[0,0]+R[1,0]*R[1,0])
    sgl = ss*arctan2d(-R[2,0],cossgl)
    om = arctan2d(R[1,0]/cossgl, R[0,0]/cossgl)
    sgu = arctan2d(R[2,1]/cossgl, R[2,2]/cossgl)
    
    QC = tasReflectionToQC(qe, UB)
    q = np.linalg.norm(QC);
    
    Ei = qe[3]
    Ef = qe[4]
    
    ki = np.sqrt(Ei)*factorsqrtEK
    kf = np.sqrt(Ef)*factorsqrtEK
    
    cos2t =(ki**2 + kf**2 - q**2) / (2. * np.abs(ki) * np.abs(kf))
    
    A4 = ss*arccosd(cos2t)
    theta = calcTheta(ki, kf, A4)
    A3 = om + ss*theta + A3Off
    A3 = np.mod(A3 + ss*180.0,360.0) - ss*180.0
    
    return A3,A4,sgu,sgl,A3+np.sign(A4)*ss*theta
    

def calcTasMisalignment(UB,planeNormal,qe):
    R = buildRMatrix(UB,planeNormal, qe[:3])
    om = arctan2d(R[1,0], R[0,0])
    return om


def calcTasQH(UBINV,angles,Ei,Ef,A3Off=0):

  A3,A4 = angles
  r = [0,0,0,A3+A3Off,A4,0.0,0.0,Ei,Ef]
  
  ki = np.sqrt(Ei)*factorsqrtEK
  kf = np.sqrt(Ef)*factorsqrtEK
  
  QV = calcTasUVectorFromAngles(r);
  
  q = np.sqrt(ki**2 +kf**2-
           2. *ki *kf * cosd(A4));
  
  if len(QV.shape)>2: # If multidimensional QV
    q = np.expand_dims(np.swapaxes(q,0,2),3)

  QV *=q
  
  Q = np.einsum('ij,...j->...i',UBINV,QV)
  
  if len(QV.shape)>2:
    QxQy = np.swapaxes(QV,0,2)
    return Q,QxQy[:,:,:,0],QxQy[:,:,:,1]

  return Q,QV

def addAuxReflection(cell,r1,r2,ss=1.0):
    h,k,l = r2[:3]#np.array([0,0,1])
    
    Ei,Ef = r1[-2:]
    
    ki,kf = np.sqrt([Ei,Ef])*factorsqrtEK
    
    B = calculateBMatrix(cell)
    
    #qe = np.array([h,k,l,Ei,Ef])
    
    
    stt1 = r1[4]
    ### makeAuxReflection
    theta = calcTheta(ki,kf,stt1)
    om = r1[3]-ss*theta
    

    om+= tasAngleBetweenReflections(B,r1[:3],[h,k,l])
    
    
    QC = np.dot(B,[h,k,l])
    q = np.linalg.norm(QC)
    
    cos2t = (ki * ki + kf * kf - q * q) /\
                (2. * np.abs(ki) * np.abs(kf))
    r2stt = np.rad2deg(np.arccos(cos2t))
    
    r2A3 = om+ss*theta
    r2A3 = np.mod(r2A3+ss*180.0,360.0)-ss*180.0
    
    r2 = np.array([h,k,l,r2A3,r2stt,0.0,0.0,Ei,Ef])

    return r2

def calTwoTheta(B,qe,ss):

  h,k,l,Ei,Ef = qe
  ki,kf = np.sqrt([Ei,Ef])*factorsqrtEK
  QC = np.dot(B,qe[:3])
  
  q = np.linalg.norm(QC);
 
  cos2t = (ki * ki + kf * kf - q * q) /\
              (2. * np.abs(ki) * np.abs(kf))
      
  return ss * np.rad2deg(np.arccos(cos2t))
 
def tasAngleBetweenReflections(B,r1,r2):
  ## Angle between the two reflections
  H1 = r1[:3]
  H2 = r2[:3]
  chi1 = np.dot(B,H1)
  chi2 = np.dot(B,H2)
  angle = np.rad2deg(np.arccos(np.dot(chi1,chi2)/(np.linalg.norm(chi1)*np.linalg.norm(chi2))))
  return angle


def calcCell(cell):
    """Calculate reciprocal lattice vectors from real space cell parameters"""

    a1,a2,a3,alpha1,alpha2,alpha3 = cell
    realVectorA = np.array([a1,0,0])
    realVectorB = a2*np.array([cosd(alpha3),sind(alpha3),0.0])#np.dot(np.array([b,0,0]),rotationMatrix(0,0,gamma))
    realVectorC = a3*np.array([cosd(alpha2),(cosd(alpha1)-cosd(alpha2)*cosd(alpha3))/sind(alpha3),
    np.sqrt(1-cosd(alpha2)**2-((cosd(alpha1)-cosd(alpha2)*cosd(alpha3))/sind(alpha3))**2)])#np.dot(np.array([c,0,0]),rotationMatrix(0,beta,0))

    volume = np.abs(np.dot(realVectorA,np.cross(realVectorB,realVectorC)))
    reciprocalVectorA = 2*np.pi*np.cross(realVectorB,realVectorC)/volume
    reciprocalVectorB = 2*np.pi*np.cross(realVectorC,realVectorA)/volume
    reciprocalVectorC = 2*np.pi*np.cross(realVectorA,realVectorB)/volume


    bv1,bv2,bv3 = reciprocalVectorA,reciprocalVectorB,reciprocalVectorC
    

    b1,b2,b3 = [np.linalg.norm(x) for x in [bv1,bv2,bv3]]
    beta1 = np.rad2deg(_tools.vectorAngle(bv2,bv3))
    beta2 = np.rad2deg(_tools.vectorAngle(bv3,bv1))
    beta3 = np.rad2deg(_tools.vectorAngle(bv1,bv2))
    cell = [a1,a2,a3,b1,b2,b3,alpha1,alpha2,alpha3,beta1,beta2,beta3]
    return cell



def converterToA3A4(Qx,Qy,Qz, Ei,Ef,A3Off=0.0,A4Sign=1): # pragma: no cover
    ## home made function to calculate A3 and A4
    Qx = np.asarray(Qx)
    Qy = np.asarray(Qy)
    Qz = np.asarray(Qz)

    QC = np.array([Qx,Qy,Qz])
    q = np.linalg.norm(QC)

    U1V = np.array([Qx.flatten(),Qy.flatten(),Qz.flatten()],dtype=float).flatten()

    U1V/=np.linalg.norm(U1V)
    U2V = np.array([0.0,0.0,1.0],dtype=float)
    
    
    TV = buildTVMatrix(U1V, U2V)
    R = np.linalg.inv(TV)
    
    ss = 1.0
    
    cossgl = np.sqrt(R[0,0]*R[0,0]+R[1,0]*R[1,0])
    om = arctan2d(R[1,0]/cossgl, R[0,0]/cossgl)
    
    ki = np.sqrt(Ei)*factorsqrtEK
    kf = np.sqrt(Ef)*factorsqrtEK
    
    cos2t =(ki**2 + kf**2 - q**2) / (2. * np.abs(ki) * np.abs(kf))
    
    A4 = ss*arccosd(cos2t)
    theta = calcTheta(ki, kf, A4)
    A3 = -om + np.sign(A4Sign)*ss*theta + A3Off
    return A3,np.sign(A4Sign)*A4


import sys
# sys.path.append(r'C:\Users\lass_j\Documents\Software\DMCpy')
from Tutorial_Class import Tutorial
import os

def Tester():
    from DMCpy import DataFile,DataSet, _tools
    import numpy as np
    
    # Create a DataFile and DataSet for 565
    file = r'data\dmc2021n000565.hdf'
    
    df = DataFile.loadDataFile(file)
    ds = DataSet.DataSet(df)
    
    
    # Create a DataFile and DataSet for 565 with correct twoTheta
    twoThetaOffset = 18.0
    
    df = DataFile.loadDataFile(file,twoThetaPosition=twoThetaOffset)
    ds = DataSet.DataSet(df)
    
    # Create a DataFile and DataSet with _tools.fileListGenerator
    scanNumbers = '578'
    folder = 'data'

    # Create complete filepath
    file = os.path.join(os.getcwd(),_tools.fileListGenerator(scanNumbers,folder)[0]) 

    df = DataFile.loadDataFile(file,twoThetaPosition=twoThetaOffset)
    ds = DataSet.DataSet(df)
    
    # If we want to load several DataFiles in the DataSet
    dataFiles = [DataFile.loadDataFile(dFP,twoThetaPosition=twoThetaOffset) for dFP in _tools.fileListGenerator(scanNumbers,folder)]
    
    ds = DataSet.DataSet(dataFiles)

    
title = 'Make a DataFile and DataSet'

introText = 'In this tutorial we demonstrate how to make DataFiles and DataSets in DMCpy. You create a DataFile by the loadDataFile function.'\
+ ' The input for loadDataFile is file name and path as a string. In addition, can arguments that act on the DataFile be give, *e.g.*  *twoThetaPosition*.'\
+ ' The _tools.fileListGenerator is a useful tool to create a list of DataFiles with complete path. The input is the short number of the Datafile and path.'\
+ ' Several DataFiles can be added into one DataSet. This is done by giving DataSet a list of DataFiles: DataSet(dataFiles)'\

outroText = ''

introText = title+'\n'+'^'*len(title)+'\n'+introText


    
Example = Tutorial('Make Data File',introText,outroText,Tester,fileLocation = os.path.join(os.getcwd(),r'docs/Tutorials'))

def test_Make_Data_File():
    Example.test()

if __name__ == '__main__':
    Example.generateTutorial()


import matplotlib.pyplot as plt
from matplotlib.widgets import Slider
import numpy as np


class InteractiveViewer(object):
    """Interactive viewer for 2D detector and 1D scan variable"""
    def __init__(self,data,twoTheta,pixelPosition,scanValues=None,scanParameter=None,
                 scanValueFormat=None,scanValueUnit=None,colorbar=False,outputFunction=print,
                 mainTitle='Single Step',vmin=None,vmax=None,
                 dataLabel = 'Intensity',axis_1_label='Sum over z',axis_2_label='Sum over 2Theta',
                 xlabel='2Theta [deg]',ylabel='z [cm]',cmap='viridis'):
        """

        args:
            - data (array): 3D array with data data
            
            - twoTheta (array): 2D array holding the 2theta values
            
            - pixelPosition (array): 3D array holding pixel position (x,y,z as function of 2theta and out of plane position)
            
        kwags:
            
            - scanValues (array): 1D array with values of scan parameter (default None)
            
            - scanParameter (str): Name of scanned parameter (default None)
            
            - scanValueFormat (str): Format string for scan parameter values (default None)
            
            - scanValueUnit (str): Unit of scan parameter (default None)
            
            - colorbar (bool): Boolean value for showing color bar (default False)
            
            - outputFunction (function): Function called when clicking on data axis (default print)

            - mainTitle (str): Title of data figure (default 'Single Step'')
                                                          
            - vmin (float): Minimum value for color plot (default None - auto)                                              
            
            - vmax (float): Maximum value for color plot (default None - auto)                                              

            - dataLabel (str): Label of the data axis (default Intensity)
            
            - axis_1_label (str): Label of axis 1 (default Sum over z)
            
            - axis_2_label (str): Label of axis 1 (default Sum over 2theta)
            
            - xlabel (str): Label of x axis (default '2Theta [deg]')
            
            - ylabel (str): Label of y axis (default 'z [cm]')

            - cmap (str): Name of color map (default viridis)

        """
        
        # Initialize index to -1 to ensure plotting of first data
        self.index = -1
        self.data = data.transpose(0,2,1) # Transpose for quicker data plotting
        
        # If scan values are not provided, create [0,1,2,3,...]
        if not scanValues is None:
            self.scanValues = scanValues
        else:
            self.scanValues = np.arange(len(data))
            
        self.dataLabel = dataLabel
        self.axis_1_label = axis_1_label
        self.axis_2_label = axis_2_label
            
        self.cmap = cmap
        self.colorbar = colorbar
        self.mainTitle = mainTitle
        
        self.xlabel = xlabel
        self.ylabel = ylabel
        
        if scanValueFormat is None: # Automatically create slider text format
            if len(self.scanValues) == 1:
                minStep = 1
            else:
                minStep = np.diff(self.scanValues).mean()
            order = int(np.ceil(np.log10(minStep)))
            decimals = np.max([0,3-order])
            self.scanValueFormat  = '{:.'+str(decimals)+'f}'#'{:.'+str(order)+'}'
            self.valfmt = '%.'+str(decimals)+'f'
            if not scanValueUnit is None:
                self.scanValueFormat = self.scanValueFormat + ' ['+scanValueUnit+']'
                self.valfmt = self.valfmt + ' ['+scanValueUnit+']'
            self.scanValueFormat = self.scanValueFormat + ' (step {})'
        # Copy two theta, and create array holding edges between values, including outer most values (length is 1 larger than twoTheta)
        self.twoTheta = twoTheta
        self.twoThetaStep = np.mean(np.diff(twoTheta[0,:]))
        
        self.twoThetaExtended = np.arange(self.twoTheta[0,0]-self.twoThetaStep*0.5,self.twoTheta[0,-1]+self.twoThetaStep*0.6,self.twoThetaStep)
        
        # Repeat for pixel position
        self.pixelPosition = pixelPosition[2]
        self.pixelPositionStep = np.mean(np.diff(self.pixelPosition[:,0]))
        self.pixelPositionExtended = np.arange(self.pixelPosition[0,0]-self.pixelPositionStep*0.5,self.pixelPosition[-1,0]+self.pixelPositionStep*0.6,self.pixelPositionStep)
        
        
        self.scanSteps = len(self.scanValues)
        self.scanValuesStep = minStep
        self.scanValuesExtended = np.arange(self.scanValues[0]-self.scanValuesStep*0.5,self.scanValues[-1]+self.scanValuesStep*0.6,self.scanValuesStep)
        
        self.scanStepExtended = np.arange(-0.5,len(self.data)-1+0.6,1.0)
        
        # Create figure with axes. Layout is 3 x 6
        self.heights = [8,1,6] # make slider thin
        self.widths = [1,3,3,3,3,1] # Change widths
        self.fig = plt.figure(constrained_layout=False,figsize=(13,9))
        
        # Create gridspec
        self.gs = self.fig.add_gridspec(3, 6,width_ratios=self.widths,height_ratios=self.heights)
        self.ax_singleStep = self.fig.add_subplot(self.gs[0, 1:-1])
        
        self.ax_singleStep.set_title(self.mainTitle)
        self.ax_slider = self.fig.add_subplot(self.gs[1, 1:-1])
        self.ax_alphaIntegrated = self.fig.add_subplot(self.gs[2,0:3])
        self.ax_alphaIntegrated.set_title(self.axis_1_label)
        # Share the yaxis between ax_alphaIntegrated and ax_thetaIntegrated
        self.ax_thetaIntegrated = self.fig.add_subplot(self.gs[2,3:],sharey=self.ax_alphaIntegrated)
        self.ax_thetaIntegrated.set_title(self.axis_2_label)
        
        
        if not scanParameter is None:
            scanLabel = scanParameter
        else:
            scanLabel = 'Scan Step'
        
        # Create slider with values between 0 and scanSteps -1)
        if self.scanSteps == 1:
            self.indexSlider = Slider(self.ax_slider, label=scanLabel, valmin=-0.5, valmax=0.5, valinit=0,valfmt=self.valfmt)
        else:
            self.indexSlider = Slider(self.ax_slider, label=scanLabel, valmin=0, valmax=self.scanSteps-1, valinit=-1,valfmt=self.valfmt)
        
        self.indexSlider.on_changed(lambda val: self.sliders_on_changed(val))
        
        
        # add labels
        self.ax_singleStep.set_xlabel(self.xlabel)
        self.ax_singleStep.set_ylabel(self.ylabel)
        
        self.ax_alphaIntegrated.set_ylabel('Scan Step')
        self.ax_alphaIntegrated.set_xlabel(self.xlabel)
        self.ax_thetaIntegrated.set_xlabel(self.ylabel)
        
        # Sum over two theta and alpha(out of plane)
        self.IThetaIntegrated = self.data.sum(axis=1)
        self.IAlphaIntegrated = self.data.sum(axis=2)
        
        if vmin is None:
            vmin = np.nanmin(self.data)
        if vmax is None:
            vmax = np.nanmax(self.data)
            
        self.initialLimits = [vmin,vmax]
        
        # Plotting limits across all steps
        
        # plot theta and alpha integrated intensities

        self.ax_alphaIntegrated._pcolormesh = self.ax_alphaIntegrated.pcolormesh(self.twoThetaExtended,self.scanStepExtended,self.IAlphaIntegrated,cmap=self.cmap)#,vmin=vmin,vmax=vmax)
        self.ax_thetaIntegrated._pcolormesh = self.ax_thetaIntegrated.pcolormesh(self.pixelPositionExtended,self.scanStepExtended,self.IThetaIntegrated,cmap=self.cmap)#,vmin=vmin,vmax=vmax)
        
        #self.ax_alphaIntegrated.axis('auto')
        #self.ax_thetaIntegrated.axis('auto')
        #self.ax_thetaIntegrated.set_yticks([])
        #self.ax_thetaIntegrated.sharey(self.ax_alphaIntegrated)
        
        
        if not scanValues is None: # if scan values are provided
            # Linear interpolation between scan value and scan index + vice versa
            # Using exteded range as to be sure to take into account the case of only as single scan step
            self.toScanValue = lambda x: np.interp(x,np.arange(-0.5,len(self.scanStepExtended)-1),self.scanValuesExtended)
            self.fromScanValue = lambda x: np.interp(x,self.scanValuesExtended,np.arange(-0.5,len(self.scanStepExtended)-1))
        
            # Create additional axes
            self.secax_theta = self.ax_thetaIntegrated.secondary_yaxis('right', functions=(self.toScanValue, self.fromScanValue))
            self.secax_alpha = self.ax_alphaIntegrated.secondary_yaxis('right', functions=(self.toScanValue, self.fromScanValue))
            ylabel = scanLabel
            if not scanValueUnit is None:
                ylabel = ylabel+' ['+scanValueUnit+']'
            self.secax_theta.set_ylabel(ylabel)
            self.secax_alpha.set_ylabel(ylabel)
        
        
        ## Use scan values in stead of index
        def formatter(value): 
            return self.scanValueFormat.format(self.toScanValue(value),int(np.round(value)))
        self.indexSlider._format = formatter 

        if self.scanSteps !=1: # force a redraw of the slider
            self.indexSlider.set_val(0)
            #self.indexSlider.set_val(0)
        
        self.fig.canvas.mpl_connect('key_press_event',lambda event: self.onkeypress(event) )
        self.fig.canvas.mpl_connect('scroll_event',lambda event: self.onscroll(event))
        # Force plot of spectrum with index 0
        self.plotSpectrum(index=0)
        
        self.cid = self.fig.canvas.mpl_connect('button_press_event', lambda event: self.onclick(event,outputFunction=outputFunction))
        
        
        ## Create format coords for summed axes
        
        self.yformat = scanLabel+"={:.3f}"
        if not scanValueUnit is None:
            self.yformat = self.yformat +' ['+scanValueUnit+']'
        
        if len(self.xlabel.split(' [')) == 2:
            alphaSumfmt = ' = {:.3f} ['.join(self.xlabel.split(' ['))+ ' '
        else:
            alphaSumfmt = self.xlabel + ' = {:f} '

        if len(self.xlabel.split(' [')) == 2:
            twoThetaSumfmt = ' = {:.3f} ['.join(self.ylabel.split(' [')) + ' '
        else:
            twoThetaSumfmt = self.ylabel + ' = {:.3f} '
            
        self.format_coord_alphaSum = lambda x,y: (alphaSumfmt + self.yformat + ' scanStep={:.0f}' ).format(x,(self.toScanValue(y)),int(np.round(y)))
        self.format_coord_twoThetaSum = lambda x,y: (twoThetaSumfmt + self.yformat + ' scanStep={:.0f}').format(x,(self.toScanValue(y)),int(np.round(y)))

        self.format_coord_data = lambda x,y: (alphaSumfmt + twoThetaSumfmt).format(x,y)
        
        self.fig.tight_layout()
       
    def plotSpectrum(self,index=0,fromSlider=False):
        
        if self.index == index or index < 0 or index >= self.scanSteps:
            return
                    

        if not self.indexSlider.val == index: # If clicked on slider a float is provided. Reset to int
            self.indexSlider.set_val(index)
            return
        
        vmin,vmax = self.initialLimits
        #extent = np.array([[f(dat) for f in [np.nanmin,np.nanmax]] for dat in [ax.twoTheta[index],ax.pixelPosition[index]]]).flatten()
        
        if hasattr(self.ax_singleStep,'_pcolormesh'):
            self.ax_singleStep._pcolormesh.set_array(self.data[index].T)
            self.ax_singleStep.redraw_in_frame()
        else:
            self.ax_singleStep._pcolormesh = self.ax_singleStep.pcolormesh(self.twoThetaExtended,self.pixelPositionExtended,self.data[index].T,vmin=vmin,vmax=vmax,cmap=self.cmap)
            plt.draw()
    
        
        self.index = index
        if self.colorbar:
            if not hasattr(self,'_colorbar'):
                self.colorbarax = self.fig.add_subplot(self.gs[0, -1])
                self._colorbar = plt.colorbar(self.ax_singleStep._pcolormesh,cax=self.colorbarax)
                self._colorbar.set_label(self.dataLabel)
            
        
    
    def increaseAxis(self,step=1): # Call function to increase index
        index = self.index
        index+=step
        if index>=len(self.data):
            index = len(self.data)-1
        self.plotSpectrum(index)
    
    def decreaseAxis(self,step=1): # Call function to decrease index
        index = self.index
        index-=step
        if index<=-1:
            index = 0
        self.plotSpectrum(index)
        
        
    def onkeypress(self,event): # pragma: no cover
        if event.key in ['+','up']:
            self.increaseAxis()
        elif event.key in ['-','down']:
            self.decreaseAxis()
        elif event.key in ['home']:
            index = 0
            self.plotSpectrum(index)
        elif event.key in ['end']:
            index = len(self.data)-1
            self.plotSpectrum(index)
        elif event.key in ['pageup','ctrl++','ctrl+up']: # Pressing pageup or page down performs steps of 10
            self.increaseAxis(step=10)
        elif event.key in ['pagedown','ctrl+-','ctrl+down']:
            self.decreaseAxis(step=10)
            
    def sliders_on_changed(self,index):
        index = int(round(index))
        if index != self.index:
            self.plotSpectrum(index,fromSlider=True)
        
    def onscroll(self,event): # pragma: no cover
        if(event.button=='up'):
            self.increaseAxis()
        elif event.button=='down':
            self.decreaseAxis()
            
    def onclick(self,event,outputFunction=print):
        
        #outputFunction(event,event.inaxes)
        if not event.inaxes is None:
            self.event = event
            if event.inaxes in [self.ax_alphaIntegrated,self.ax_thetaIntegrated]:
                if event.button != 1: # not a left click
                    return
                # Check if anything but no special curser is chosen
                try:
                    C = self.fig.canvas.cursor().shape() # Only works for pyQt5 backend
                except:
                    pass
                else:
                    if C != 0:
                        return
                ## Plot corresponding scan step
                self.plotSpectrum(index = int(np.round(self.event.ydata)))
            elif event.inaxes == self.ax_singleStep:
                position = event.xdata,event.ydata
                outputFunction(event.inaxes.format_coord(*position))
            
        
        #### Functions to interface with format coord of axes with data
    def format_coord_alphaSum_getter(self):
        return self.ax_alphaIntegrated.format_coord
        
    def format_coord_alphaSum_setter(self,function):
        self.ax_alphaIntegrated.format_coord = function
        
    def format_coord_twoThetaSum_getter(self):
        return self.ax_thetaIntegrated.format_coord
        
    def format_coord_twoThetaSum_setter(self,function):
        self.ax_thetaIntegrated.format_coord = function
    
    def format_coord_data_getter(self):
        return self.ax_singleStep.format_coord
        
    def format_coord_data_setter(self,function):
        self.ax_singleStep.format_coord = function
        
    format_coord_alphaSum = property(format_coord_alphaSum_getter, format_coord_alphaSum_setter)
    format_coord_twoThetaSum = property(format_coord_twoThetaSum_getter, format_coord_twoThetaSum_setter)
    format_coord_data = property(format_coord_data_getter, format_coord_data_setter)
    

    def set_all_clim(self,vmin,vmax):
        for pc in [self.ax_alphaIntegrated,self.ax_thetaIntegrated,self.ax_singleStep]:
            pc._pcolormesh.set_clim(vmin,vmax)
    
    def set_clim(self,vmin,vmax):
        self.ax_singleStep._pcolormesh.set_clim(vmin,vmax)


import sys
# sys.path.append(r'C:\Users\lass_j\Documents\Software\DMCpy')
from Tutorial_Class import Tutorial
import os

def Tester():
    from DMCpy import DataFile, _tools
    
    # Give file number and folder the file is stored in.
    scanNumbers = '578'
    folder = 'data'
    
    # Give correct two theta
    twoThetaOffset = 18.0
    
    # Create complete filepath
    file = os.path.join(os.getcwd(),_tools.fileListGenerator(scanNumbers,folder)[0]) 

    # Load data file with corrected twoTheta
    df = DataFile.loadDataFile(file,twoThetaPosition=twoThetaOffset)
    
    #Plot detector
    ax = df.plotDetector()
    
    fig = ax.get_figure()
    fig.set_size_inches(20, 2.5)
    fig.tight_layout()
    fig.savefig(r'docs/Tutorials/Powder/Plot2DPowderDetector_new_and_better.png',format='png',dpi=300)
    
title = 'Detector Overview Powder'

introText = 'The simplets data set on the DMC instrument is that of a powder measured with only one setting. This results '\
   + 'in a \'one shot\' data set where scattering intensity is measured as a function of scattering angle and position '\
   + 'out of plane. This can be visualized in the frame of reference of the instrument by the following code:'


outroText = 'At the current stage, a normalization file for the 2D detector is not present and thus a dummy is created. '\
    +'Running the above code generates the following images showing neutron intensity as function of 2Theta and out of plane position:'\
+'\n .. figure:: Plot2DPowderDetector_new_and_better.png\n  :width: 100%\n  :align: center\n\n '

introText = title+'\n'+'^'*len(title)+'\n'+introText


    
Example = Tutorial('2D Detector Plot',introText,outroText,Tester,fileLocation = os.path.join(os.getcwd(),r'docs/Tutorials/Powder'))

def test_2D_Detector_Plot():
    Example.test()

if __name__ == '__main__':
    Example.generateTutorial()


from setuptools import setup
import setuptools
import os
import sys

_here = os.path.abspath(os.path.dirname(__file__))
operatingSystem = sys.platform


with open(os.path.join(_here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


installFolder = os.path.abspath(os.path.join(os.path.split(setuptools.__file__)[0],'..'))
pythonPath =  os.path.relpath(os.path.join(installFolder,'DMCpy'),sys.base_prefix)

setup(
    name='DMCpy',
    version='0.2.1',
    description=('Python software packaged designed for reduction of neutron powder diffraction data from DMC at PSI.'),
    long_description=long_description,
    author='Jakob Lass',
    author_email='jakob.lass@psi.ch',
    url='https://www.psi.ch/en/sinq/dmc/',
    license='MPL-2.0',
    data_files = [(pythonPath, ["LICENSE.txt"]),((os.path.join(pythonPath),["DMCpy/calibrationDict.dat"]))],
    packages=['DMCpy','DMCpy/CommandLineScripts'],
    entry_points = {
        "console_scripts": []#'MJOLNIRHistory = MJOLNIR.CommandLineScripts.MJOLNIRHistory:main',
        },
    python_requires='>=3.6',
    install_requires=['matplotlib>=3','numpy>=1.14','h5py>=2.5','scipy','datetime','pandas','future','crystals',
                    'pip>=20'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'],
    )



import functools
import numpy as np
from difflib import SequenceMatcher
import os.path
import cProfile, pstats, io
from itertools import product


MPLKwargs = ['agg_filter','alpha','animated','antialiased or aa','clip_box','clip_on','clip_path','color or c','contains','dash_capstyle','dash_joinstyle','dashes','drawstyle','figure','fillstyle','gid','label','linestyle or ls','linewidth or lw','marker','markeredgecolor or mec','markeredgewidth or mew','markerfacecolor or mfc','markerfacecoloralt or mfcalt','markersize or ms','markevery','path_effects','picker','pickradius','rasterized','sketch_params','snap','solid_capstyle','solid_joinstyle','transform','url','visible','xdata','ydata','zorder']

def KwargChecker(function=None,include=None):
    """Function to check if given key-word is in the list of accepted Kwargs. If not directly therein, checks capitalization. If still not match raises error
    with suggestion of closest argument.
    
    Args:
    
        - func (function): Function to be decorated.

    Raises:

        - AttributeError
    """
    def KwargCheckerNone(func):
        @functools.wraps(func)
        def newFunc(*args,**kwargs):
            argList = extractArgsList(func,newFunc,function,include)
            checkArgumentList(argList,kwargs)
            returnval = func(*args,**kwargs)
            return returnval
        newFunc._original = func
        newFunc._include = include
        newFunc._function = function
        return newFunc
    return KwargCheckerNone

def extractArgsList(func,newFunc,function,include):
    N = func.__code__.co_argcount # Number of arguments with which the function is called
    argList = list(newFunc._original.__code__.co_varnames[:N]) # List of arguments
    if not function is None:
        if isinstance(function,(list,np.ndarray)): # allow function kwarg to be list or ndarray
            for f in function:
                for arg in f.__code__.co_varnames[:f.__code__.co_argcount]: # extract all arguments from function
                    argList.append(str(arg))
        else: # if single function
            for arg in function.__code__.co_varnames[:function.__code__.co_argcount]:
                argList.append(str(arg))
    if not include is None:
        if isinstance(include,(list,np.ndarray)):
            for arg in include:
                argList.append(str(arg))
        else:
            argList.append(str(include))
        argList = list(set(argList)) # Cast to set to remove duplicates
        argList.sort() #  Sort alphabetically
    return argList

def checkArgumentList(argList,kwargs):
    notFound = []
    for key in kwargs:
        if key not in argList:
            similarity = np.array([SequenceMatcher(None, key.lower(), x.lower()).ratio() for x in argList])
            maxVal = np.max(similarity)
            maxId = np.argmax(similarity)
            notFound.append('Key-word argument "{}" not understood. Did you mean "{}"?'.format(key,argList[maxId]))
    if len(notFound)>0:
        if len(notFound)>1:
            errorMsg = 'The following key-word arguments are not understood:\n'
            errorMsg+='\n'.join(notFound)
        else:
            errorMsg = notFound[0]
        error = AttributeError(errorMsg)
        raise error

@KwargChecker()
def numberStringGenerator(fileNames,instrumentName='dmc'):
    names = np.array([os.path.splitext(os.path.basename(df))[0] for df in fileNames])
    # Find base name and remove extension
    if len(fileNames) != 1:
        prefix = os.path.commonprefix(list(names))
        
        if instrumentName in prefix:
            # Remove all non-zero digits from prefix
            while prefix[-1]!='0' and prefix[-1]!='n':
                prefix = prefix[:-1]
            year = int(prefix[len(instrumentName):len(instrumentName)+4])
            numbers = np.array([n[len(prefix):] for n in names],dtype=int)
            sortNumbers = np.sort(numbers)
            diff = np.diff(sortNumbers)
            separators = list(np.arange(len(diff))[diff>1]+1) # add one due to diff removing 1 lenght
            groups = []
            if len(separators) == 0:
                groups.append('-'.join([str(sortNumbers[0]),str(sortNumbers[-1])]))
            else:
                separators.insert(0,0)
                separators.append(-1)
                for start,stop in zip(separators[:-1],separators[1:]):
                    if stop == -1:
                        group = sortNumbers[start:]
                    else:
                        group = sortNumbers[start:stop]
                    if len(group)>2:
                        groups.append('-'.join([str(group[0]),str(group[-1])]))
                    elif len(group)==2:
                        groups.append(','.join(group.astype(str)))
                    else:
                        groups.append(str(group[0]))
            files = ','.join(groups)
    else:
        splitting = fileNames[0].split('n')
        InstrumentYear = 'n'.join(splitting[:-1])
        fileNumbers = splitting[-1]
        files = str(int(fileNumbers.split('.')[0]))
        year = int(InstrumentYear[-4:])
    return year,files

@KwargChecker()
def fileListGenerator(numberString,folder,year=2021, format = None, instrument = 'dmc'):
    """Function to generate list of data files.
    
    Args:
        
        - numberString (str): List if numbers separated with comma and dashes for sequences.
        
        - folder (str): Folder of wanted data files.
        
    Kwargs:

        - year (int): Year of wanted data files (default 2018)

        - format (str): format of data files (default None, but dmc if instrument is provided)

        - instrument (str): Instrument to be used to determine format string (default dmc)
        
    returns:
        
        - list of strings: List containing the full file string for each number provided.
        
    Example:
        >>> numberString = '201-205,207-208,210,212'
        >>> files = fileListGenerator(numberString,'data/',2018)
        ['data/dmc2018n000201.hdf', 'data/dmc2018n000202.hdf', 
        'data/dmc2018n000203.hdf', 'data/dmc2018n000204.hdf', 
        'data/dmc2018n000205.hdf', 'data/dmc2018n000207.hdf', 
        'data/dmc2018n000208.hdf', 'data/dmc2018n000210.hdf', 
        'data/dmc2018n000212.hdf']
    """
        
    splits = numberString.split(',')
    dataFiles = []
    if format is None: # If no user specified format is provided
        if instrument == 'dmc':
            format = 'dmc{:d}n{:06d}.hdf'
        else:
            raise AttributeError('Provided instrument "{}" not understood'.format(instrument))


    for sp in splits:
        isRange = sp.find('-')!=-1
        
        if isRange:
            spSplits = sp.split('-')
            if len(spSplits)>2:
                raise AttributeError('Sequence "{}" not understood - too many dashes.'.format(sp))
            startNumber = int(spSplits[0])
            endNumber = int(spSplits[1])
            numbers = np.arange(startNumber,endNumber+1)    
        else:
            numbers = [int(sp)]

        dataFiles.append([os.path.join(folder,format.format(year,x)) for x in numbers])
    return list(np.concatenate(dataFiles))

def roundPower(x,default=4):
    """Round to nearest 10^x"""
    if not np.isclose(x,0.0): # Sign to fit with np.round
        return -int(np.floor(np.log10(np.abs(x))))
        
    else:
        return default


def binData3D(dx,dy,dz,pos,data,norm=None,mon=None,bins=None):
    """ 3D binning of data.

    Args:

        - dx (float): Step size in x (required).

        - dy (float): Step size in x (required).

        - dz (float): Step size in x (required).

        - pos (2D array): Position of data points as flattened lists (X,Y,Z) (required).

        - data (array): Flattened data array (required).

    Kwargs:

        - norm (array): Flattened normalization array.

        - mon (array): Flattened monitor array.

        - bins (list of arrays): Bins locating edges in the x, y, and z directions.

    returns:

        Re-binned intensity (and if provided Normalization, Monitor, and Normalization Count) and X, Y, and Z bins in 3 3D arrays.


    Example:

    >>> pos = [Qx,Qy,E]
    >>> Data,bins = DataSet.binData3D(0.05,0.05,0.2,pos,I,norm=Norm,mon=Monitor)

    """

    if bins is None:
        bins = calculateBins(dx=dx,dy=dy,dz=dz,pos=pos)
    if len(pos[0].shape)>1: # Flatten positions
        pos = np.array([x.flatten() for x in pos])
    #NonNaNs = 1-np.isnan(data.flatten())

    #pos = [np.array(x[NonNaNs]) for x in pos]
    HistBins = [bins[0][:,0,0],bins[1][0,:,0],bins[2][0,0,:]]
    intensity =    np.histogramdd(np.array(pos).T,bins=HistBins,weights=data.flatten())[0].astype(data.dtype)

    returndata = [intensity]
    if mon is not None:
        MonitorCount=  np.histogramdd(np.array(pos).T,bins=HistBins,weights=mon.flatten())[0].astype(mon.dtype)
        returndata.append(MonitorCount)
    if norm is not None:
        Normalization= np.histogramdd(np.array(pos).T,bins=HistBins,weights=norm.flatten())[0].astype(norm.dtype)
        
        returndata.append(Normalization)
        
    NormCount =    np.histogramdd(np.array(pos).T,bins=HistBins,weights=np.ones_like(data).flatten())[0].astype(float)
    returndata.append(NormCount)
    return returndata,bins


def calculateBins(dx,dy,dz,pos):
    diffx = np.abs(np.max(pos[0])-np.min(pos[0]))
    diffy = np.abs(np.max(pos[1])-np.min(pos[1]))
    diffz = np.abs(np.max(pos[2])-np.min(pos[2]))
    
    xbins = np.round(diffx/dx).astype(int)+1
    ybins = np.round(diffy/dy).astype(int)+1
    zbins = np.round(diffz/dz).astype(int)+1
    
    _X = np.linspace(np.min(pos[0]),np.max(pos[0]),xbins)
    _Y = np.linspace(np.min(pos[1]),np.max(pos[1]),ybins)
    _Z = np.linspace(np.min(pos[2]),np.max(pos[2]),zbins)
    
    X,Y,Z = np.meshgrid(_X,_Y,_Z,indexing='ij')
    
    XX,YY,ZZ = calculateGrid3D(X=X,Y=Y,Z=Z)
    
    bins=[XX,YY,ZZ]
    return bins



def calculateGrid3D(X,Y,Z):
    """Generate 3D grid with centers given by X,Y, and Z.
     Args:
        
        X (3D array): 3D array of x values generated by np.meshgrid.
                
        Y (3D array): 3D array of y values generated by np.meshgrid.
                
        Z (3D array): 3D array of z values generated by np.meshgrid.
        
    Example:

    >>> x = np.linspace(-1.5,1.5,20)
    >>> y = np.linspace(0,1.5,10)
    >>> z = np.linspace(-1.0,5.5,66)
    >>> X,Y,Z = np.meshgrid(x,y,z,indexing='ij')
    >>> XX,YY,ZZ = calculateGrid3D(X,Y,Z)

    Now XX is a 21x11x67 array containing all x coordinates of the edges exactly midway between the points. Same goes for YY and ZZ with y and z coordinates respectively.
    """

    xshape = np.array(X.shape)
    if np.any(xshape <= 1):
        raise AttributeError('Provided array has dimension(s) {} of size <= 1'.format(xshape))
    XT = np.zeros((xshape[0]+1,xshape[1]+1,xshape[2]+1))
    YT = np.zeros_like(XT)
    ZT = np.zeros_like(XT)
    
    
    
    dx0 = np.diff(X,axis=0)
    dx1 = np.diff(X,axis=1)
    dx2 = np.diff(X,axis=2)
    dy0 = np.diff(Y,axis=0)
    dy1 = np.diff(Y,axis=1)
    dy2 = np.diff(Y,axis=2)
    dz0 = np.diff(Z,axis=0)
    dz1 = np.diff(Z,axis=1)
    dz2 = np.diff(Z,axis=2)
    
    
    XX = X.copy()
    XX[:-1]-=0.5*dx0
    XX[-1]-=0.5*dx0[-1]
    XX[:,:-1]-=0.5*dx1
    XX[:,-1]-=0.5*dx1[:,-1]
    XX[:,:,:-1]-=0.5*dx2
    XX[:,:,-1]-=0.5*dx2[:,:,-1]
    
    YY = Y.copy()
    YY[:-1]-=0.5*dy0
    YY[-1]-=0.5*dy0[-1]
    YY[:,:-1]-=0.5*dy1
    YY[:,-1]-=0.5*dy1[:,-1]
    YY[:,:,:-1]-=0.5*dy2
    YY[:,:,-1]-=0.5*dy2[:,:,-1]
    
    ZZ = Z.copy()
    ZZ[:-1]-=0.5*dz0
    ZZ[-1]-=0.5*dz0[-1]
    ZZ[:,:-1]-=0.5*dz1
    ZZ[:,-1]-=0.5*dz1[:,-1]
    ZZ[:,:,:-1]-=0.5*dz2
    ZZ[:,:,-1]-=0.5*dz2[:,:,-1]
    
    XT[:-1,:-1,:-1]=XX.copy()
    YT[:-1,:-1,:-1]=YY.copy()
    ZT[:-1,:-1,:-1]=ZZ.copy()
    
    
    XT[-1,:-1,:-1]=XT[-2,:-1,:-1]+dx0[-1]
    XT[:-1,-1,:-1]=XT[:-1,-2,:-1]+dx1[:,-1,:]
    XT[:-1,:-1,-1]=XT[:-1,:-1,-2]+dx2[:,:,-1]
    XT[:-1,-1,-1]=0.5*(XT[:-1,-1,-2]+dx2[:,-1,-1]+XT[:-1,-2,-1]+dx1[:,-1,-1])
    XT[-1,:-1,-1]=0.5*(XT[-1,:-1,-2]+dx2[-1,:,-1]+XT[-2,:-1,-1]+dx0[-1,:,-1])
    XT[-1,-1,:-1]=0.5*(XT[-1,-2,:-1]+dx1[-1,-1,:]+XT[-2,-1,:-1]+dx0[-1,-1,:])
    XT[-1,-1,-1]=(XT[-1,-2,-1]+dx1[-1,-1,-1]+XT[-2,-1,-1]+dx0[-1,-1,-1]+XT[-1,-1,-2]+dx2[-1,-1,-1])/3
    
    YT[-1,:-1,:-1]=YT[-2,:-1,:-1]+dy0[-1]
    YT[:-1,-1,:-1]=YT[:-1,-2,:-1]+dy1[:,-1,:]
    YT[:-1,:-1,-1]=YT[:-1,:-1,-2]+dy2[:,:,-1]
    YT[:-1,-1,-1]=0.5*(YT[:-1,-1,-2]+dy2[:,-1,-1]+YT[:-1,-2,-1]+dy1[:,-1,-1])
    YT[-1,:-1,-1]=0.5*(YT[-1,:-1,-2]+dy2[-1,:,-1]+YT[-2,:-1,-1]+dy0[-1,:,-1])
    YT[-1,-1,:-1]=0.5*(YT[-1,-2,:-1]+dy1[-1,-1,:]+YT[-2,-1,:-1]+dy0[-1,-1,:])
    YT[-1,-1,-1]=(YT[-1,-2,-1]+dy1[-1,-1,-1]+YT[-2,-1,-1]+dy0[-1,-1,-1]+YT[-1,-1,-2]+dy2[-1,-1,-1])/3
    
    ZT[-1,:-1,:-1]=ZT[-2,:-1,:-1]+dz0[-1]
    ZT[:-1,-1,:-1]=ZT[:-1,-2,:-1]+dz1[:,-1,:]
    ZT[:-1,:-1,-1]=ZT[:-1,:-1,-2]+dz2[:,:,-1]
    ZT[:-1,-1,-1]=0.5*(ZT[:-1,-1,-2]+dz2[:,-1,-1]+ZT[:-1,-2,-1]+dz1[:,-1,-1])
    ZT[-1,:-1,-1]=0.5*(ZT[-1,:-1,-2]+dz2[-1,:,-1]+ZT[-2,:-1,-1]+dz0[-1,:,-1])
    ZT[-1,-1,:-1]=0.5*(ZT[-1,-2,:-1]+dz1[-1,-1,:]+ZT[-2,-1,:-1]+dz0[-1,-1,:])
    ZT[-1,-1,-1]=(ZT[-1,-2,-1]+dz1[-1,-1,-1]+ZT[-2,-1,-1]+dz0[-1,-1,-1]+ZT[-1,-1,-2]+dz2[-1,-1,-1])/3
    
    
    return XT,YT,ZT

def rotMatrix(v,theta,deg=True):
    """ Generalized rotation matrix.
    
    Args:
        
        - v (list): Rotation axis around which matrix rotates
        
        - theta (float): Rotation angle (by default in degrees)
        
    Kwargs:
        
        - deg (bool): Whether or not angle is in degrees or radians (Default True)
        
    Returns:
        
        - 3x3 matrix rotating points around vector v by amount theta.
    """
    if deg==True:
        theta = np.deg2rad(theta.copy())
    v/=np.linalg.norm(v)
    m11 = np.cos(theta)+v[0]**2*(1-np.cos(theta))
    m12 = v[0]*v[1]*(1-np.cos(theta))-v[2]*np.sin(theta)
    m13 = v[0]*v[2]*(1-np.cos(theta))+v[1]*np.sin(theta)
    m21 = v[0]*v[1]*(1-np.cos(theta))+v[2]*np.sin(theta)
    m22 = np.cos(theta)+v[1]**2*(1-np.cos(theta))
    m23 = v[1]*v[2]*(1-np.cos(theta))-v[0]*np.sin(theta)
    m31 = v[0]*v[2]*(1-np.cos(theta))-v[1]*np.sin(theta)
    m32 = v[1]*v[2]*(1-np.cos(theta))+v[0]*np.sin(theta)
    m33 = np.cos(theta)+v[2]**2*(1-np.cos(theta))
    return np.array([[m11,m12,m13],[m21,m22,m23],[m31,m32,m33]])


def Norm2D(v):
    reciprocal = np.abs(1/v)
    if np.isclose(reciprocal[0],reciprocal[1]):
        return v*reciprocal[0]
    
    ratio = np.max(reciprocal)/np.min(reciprocal)
    if np.isclose(np.mod(ratio,1),0.0) or np.isclose(np.mod(ratio,1),1.0):
        return v*np.min(reciprocal)*ratio
    else:
        return v

def LengthOrder(v):
    nonZeroPos = np.logical_not(np.isclose(v,0.0))
    if np.sum(nonZeroPos)==1:
        Rv = v/np.linalg.norm(v)
        return Rv
    if np.sum(nonZeroPos)==0:
        raise AttributeError('Provided vector is zero vector!')
    
    if np.sum(nonZeroPos)==3:
        v1 = Norm2D(v[:2])
        ratio = v1[0]/v[0]
        v2 = Norm2D(np.array([v1[0],v[2]*ratio]))
        ratio2 = v2[0]/v1[0]
        Rv = np.array([v2[0],v1[1]*ratio2,v2[1]])
    else:
        Rv = np.zeros(3)
        nonZeros = v[nonZeroPos]
        Rv[nonZeroPos] = Norm2D(nonZeros)
    
    if not np.isclose(np.dot(Rv,v)/(np.linalg.norm(Rv)*np.linalg.norm(v)),1.0):
        raise AttributeError('The found vector is not parallel to original vector: {}, {}',format(Rv,v))
    return Rv

def overWritingFunctionDecorator(overWritingFunction):
    def overWriter(func):
        return overWritingFunction
    return overWriter


@KwargChecker()
def vectorAngle(V1,V2):
    """calculate angle between V1 and V2.
    
    Args:
    
        - V1 (list): List or array of numbers
        
        - V2 (list): List or array of numbers
        
    Return:
        
        - theta (float): Angle in degrees between the two vectors
    """
    return np.arccos(np.dot(V1,V2.T)/(np.linalg.norm(V1)*np.linalg.norm(V2)))

def normlength(V):
    """rescale V to have unit length"""
    return V/np.linalg.norm(V)


def invert(M):
    """Invert non-square matrices as described on https://en.wikipedia.org/wiki/Generalized_inverse.
    
    Args:
        
        - M (matrix): Matrix in question.
        
    Returns:
        
        - Left or right inverse matrix depending on shape of provided matrix.
    """
    s = M.shape
    if s[0]>s[1]:
        return np.dot(np.linalg.inv(np.dot(M.T,M)),M.T)
    else:
        return np.dot(M.T,np.linalg.inv(np.dot(M,M.T)))



def profile(fnc):
    
    """A decorator that uses cProfile to profile a function"""
    
    def inner(*args, **kwargs):
        
        pr = cProfile.Profile()
        pr.enable()
        retval = fnc(*args, **kwargs)
        pr.disable()
        s = io.StringIO()
        sortby = 'cumulative'
        ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
        ps.print_stats()
        print(s.getvalue())
        return retval

    return inner



class CentreOfMass(object):
    """Small helper class holding a center of gravity and its weight"""
    def __init__(self,position,weight):
        x,y,z = position
        self.x = x
        self.y = y
        self.z = z
        self.weight = weight
        self.originals = [[position,weight]]
        
    def addPoint(self,position,weight=1.0):
        totalWeight = self.weight+weight
        newPosition = np.sum([self.position*self.weight,position*weight],axis=0)/totalWeight
        self.position = newPosition
        self.weight=totalWeight
        self.originals.append([position,weight])
        
    @property
    def position(self):
        return np.array([self.x,self.y,self.z])
    
    @position.setter
    def position(self,newPosition):
        x,y,z = newPosition
        self.x = x
        self.y = y
        self.z = z
        
    def __str__(self):
        return "CentreOfMass at ({},{},{}) with weight {}".format(*self.position,self.weight)
        
def distance(a,b,dx=1,dy=1,dz=1):
    """Calculate distance with variable metric"""
    return np.linalg.norm(np.dot(a-b,[dx,dy,dz]))



def clusterPoints(positions,weights=None,distanceThreshold=0.01, shufflePoints=True, distanceFunction=None):
    """Combine positions within dinstance threshold into centres of gravity with the provided weights
    
    Args:
        
        - positions (list [n,3]): List of positions to be combined
        
    Kwargs:
        
        - weights (list [n]): Weights corresponding to positions (default None -> ones)
        
        - distanceThreshold (float): Distance within which points are to be combined (default 0.01)
        
        - shufflePoints (bool): If True, shuffle the provided positions and correspondingly their weights (default False)
        
        - distanceFunction (function): Function to calculate distance (default None -> np.linalg.norm)
        
    Returns:
        
        - 
        
    Note:
        
        In situations where the positions provided are correlated it can happen that 
        the found centres of gravity are different than expected.
        
    
    """
    positions = np.asarray(positions)
    
    if distanceFunction is None:
        distanceFunction = lambda a,b: np.linalg.norm(a-b)
    if weights is None:
        weights = np.zeros(len(positions))
    else:
        weights = np.asarray(weights)
        
    if shufflePoints:
        shuffled = np.concatenate([positions,weights.reshape(-1,1)],axis=1)
        np.random.shuffle(shuffled)
        positions = shuffled[:,:3]
        weights = shuffled[:,-1]
        
    centres = [CentreOfMass(weight=weights[0],position=positions[0])]
    
    for I,(pos,weight) in enumerate(zip(positions[1:],weights[1:])):
        #if np.mod(I,100):
            #print(I,'len(peak) = ',len(peaks))
        posUsed = False
        for p in centres:
            if distanceFunction(p.position,pos)<distanceThreshold:
                p.addPoint(pos,weight=weight)
                posUsed = True
                break
            
        if not posUsed:
            centres.append(CentreOfMass(pos,weight))
    return centres


def calculateTriplets(reflections,normalized=False):
    """Calculate cross product triples between all points
    
    Args: 
        
        - reflections (list [n,3]): List of reflections to use
        
    Kwargs:
        
        - normalized (bool): Normalize the length of normal (default False)
        
    """
    tripletNormal = []
     
    points = [np.squeeze(a) for a in np.vsplit(reflections, reflections.shape[0])]
    np.random.shuffle(points)
    # Calculate the cross product of the vectors connecting 3 random points
    
    for a1, a2, a3 in product(points, repeat=3):
        # find cross product
        normal = np.cross(a2 - a1, a3 - a1)
        
        # if length of cross product is 0, continue
        if np.allclose(normal, 0, atol=1e-2):
            continue
        if normalized:
            # make it normalized and make it 'mostly positive'
            normal /= np.linalg.norm(normal)
            normal*=np.sign(np.sum(normal))
        
        tripletNormal.append(normal)
    return tripletNormal


def plusMinusGenerator():
    """generator giving an infinite series following 0, -1, 1, -2, 2, ..."""

    yield 0
    start = 0
    while True:
        if start>0:
            start=-start
        else:
            start=-start+1
        yield start


def calculateHKLWithinQLimitsGenerator(BMatrix,QMin=0,QMax=10):
    """Generator to calculate all HKLs within a range of Qs"""
    for h in plusMinusGenerator():
        for k in plusMinusGenerator():
            for l in plusMinusGenerator():
                q = np.linalg.norm(np.dot(BMatrix,[h,k,l]))
                if (q>QMax*1.5):
                    break
                if q>QMin and q<QMax:
                    yield [h,k,l]
            if np.linalg.norm(np.dot(BMatrix,[h,k,0]))>QMax*1.5:
                break
        if np.linalg.norm(np.dot(BMatrix,[h,0,0]))>QMax*1.5:
            break
    



def calculateHKLWithinQLimits(BMatrix,QMin=0,QMax=10):
    positions = []
    
    for h in plusMinusGenerator():
        for k in plusMinusGenerator():
            for l in plusMinusGenerator():
                q = np.linalg.norm(np.dot(BMatrix,[h,k,l]))
                if (q>QMax*1.5):
                    break
                if q>QMin and q<QMax:
                    positions.append([h,k,l])
            if np.linalg.norm(np.dot(BMatrix,[h,k,0]))>QMax*1.5:
                break
        if np.linalg.norm(np.dot(BMatrix,[h,0,0]))>QMax*1.5:
            break
    return positions


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 10 10:31:10 2019

@author: lass
"""

import sys
import glob
import os

if len(sys.argv)==1:
	
	list_of_files = glob.glob('dist/*') # * means all if need specific format then *.csv
	latest_file = os.path.split(max(list_of_files, key=os.path.getctime))[-1]


	latestVersion = [int(x) for x in latest_file.split('-')[-1].replace('.tar.gz','').split('.')]

	latestVersion[-1]+=1
	version = '{}.{}.{}'.format(*latestVersion)
else:
	version = sys.argv[1]

print('Updating to version ',version)

# update to new version in setup.py
with open('setup.py') as f:
	lines = f.readlines()
	writeLines = ''
	for l in lines:
		if l.find("    version='")!=-1:
			l ="    version='"+version+"',\n"
		writeLines+=l
			
with open('setup.py','w') as f:
	f.write(writeLines)

with open('DMCpy/__init__.py') as f:
	lines = f.readlines()
	writeLines = ''
	for l in lines:
		if l.find("__version__ = ")!=-1:
			idx = l.find('__version__ = ')
			l = l[:idx] + "__version__ = '"+version+"'\n"
		writeLines+=l
        
with open('DMCpy/__init__.py','w') as f:
	f.write(writeLines)

with open('docs/conf.py') as f:
	lines = f.readlines()
	writeLines = ''
	for l in lines:
		if l.find("version = u'")!=-1:
			l = "version = u'"+version+"'\n"
		elif l.find("release = u'")!=-1:
			l = "release = u'"+version+"'\n"
		writeLines+=l
			
with open('docs/conf.py','w') as f:
	f.write(writeLines)

with open('test/init.py') as f:
	lines = f.readlines()
	writeLines = ''
	for l in lines:
		if l.find("__version__")!=-1:
			idx = l.find('__version__==')
			l = l[:idx] + "__version__=='"+version+"')\n"
		writeLines+=l
        
with open('test/init.py','w') as f:
	f.write(writeLines)
if False:
    os.system("git add DMCpy/__init__.py test/init.py setup.py")
    os.system('git commit -m "Update to version {}"'.format(version))
    os.system('git tag -a {} -m "{}"'.format(version,version))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 23 08:09:38 2019

@author: lass
"""
#import pytest

import inspect
import pytest,types
import re



import os,sys,numpy as np

def formatCode(text,indentChar = '   ', skipHeader=1):
    """Formater for raw code strings to doc string formation"""
    
    text= text.split('\n')[skipHeader:]
    
    newText = []
    newText.append('.. code-block:: python\n   :linenos:\n')
    figCounter = 0
    for line in text:
        if line[:8] == ' '*8: # Replace initial blank with >>>
            line = indentChar + line[8:]
        elif line[:4] == ' '*4:
            line = indentChar + line[4:]
        elif len(line)==0:
            line = indentChar
            
        if line.find('savefig') != -1: # code contains a save function
            startP = line.find('(')
            endP = line.find(')')
            line = line[:startP] + "('figure{}.png',format='png')".format(figCounter) + line[endP+1:]
            figCounter+=1


        if (line.find(os.path.sep)!=-1 or line.find(os.path.altsep)!=-1) and line.find('DMC')!=-1:
            matches = re.findall(r'(\/.*?\.[\w:]+)',line)
            for match in matches:
                match
                fileName = os.path.split(match)[-1]
                line = line.replace(match,os.path.join('Path','To','Data','Folder',fileName)).replace('C:','')
        
            
        newText.append(line)
            
    return '\n'.join(newText)

class Tutorial(object):
    def __init__(self,name,introText,outroText,code,fileLocation=None,dependentFiles = None):
        self.name = name
        self.introText = introText
        self.outroText = outroText
        self.code = code
        self.dependentFiles = dependentFiles
        if not fileLocation is None:
            if fileLocation[-1] != os.path.sep:
                fileLocation += os.path.sep
            self.fileLocation = os.path.join(fileLocation,self.name.replace(' ','_')+'.rst')
        else:
            self.fileLocation = fileLocation
        
    def test(self):
        # if sys.platform == 'win32':
        #     if sys.version_info.major == 3 and sys.version_info.minor <8: # replace on code objects is not supported
        #         raise NotImplementedError('Cannot run this on your current version of python. You have {} but python 3.8 or later is needed.'.format(sys.version))
        #     Tester = self.code

        #     consts = Tester.__code__.co_consts
        #     tempList = np.array(list(consts),dtype=object)
        #     returnList = []
        #     for item in tempList:
        #         try:
        #             if '/home/' in item:
        #                 if 'Software' in item:
        #                     item = item.replace('/home/lass/Dropbox/PhD/Software',r'C:/Users/lass_j/Documents/Software').replace('/','\\')
        #                 elif 'CAMEAData' in item:
        #                     item = item.replace('/home/lass/Dropbox/PhD/CAMEAData',r'C:/Users/lass_j/Documents/CAMEA2018').replace('/','\\')
        #         except:
        #             pass
        #         returnList.append(item)
        #
        #    Tester.__code__ = Tester.__code__.replace(co_consts=tuple(returnList))
        #    Tester()
        #else:
        self.code()
        
    def generateTutorial(self):

        if not self.dependentFiles is None:
            
            folder = os.path.sep.join(self.fileLocation.split(os.path.sep)[:-1])+os.path.sep
            print("Copying files to {}".format(folder))
            from shutil import copyfile
            for f in list(self.dependentFiles):
                copyfile(f, folder+f.split(os.path.sep)[-1])
        
        # Test if code is running
        codeLocation = inspect.getsourcefile(self.code)
        codeFunctionName = 'test_'+self.name.replace(' ','_')
        #print('pytest '+'{}::{}'.format(codeLocation,codeFunctionName))
        print('Running tests for "{}" in file "{}"'.format(codeFunctionName,codeLocation.split(os.path.sep)[-1]))
        result = pytest.main(args=['-q','{}::{}'.format(codeLocation,codeFunctionName)])
        
        if result != 0:
            return None
            #raise RuntimeError('An error occurred while running pytest for "{}" defined in function "{}"'.format(codeFunctionName,codeLocation))    
        else:
            print('Test successful!')

        #if sys.platform == 'win32':
        #    fileLocation = self.fileLocation.replace('/home/lass/Dropbox/PhD/Software',r'C:/Users/lass_j/Documents/Software').replace('/','\\')
        #else:
        fileLocation = self.fileLocation
        
        introText = self.introText
        outroText = self.outroText
        codeText = inspect.getsource(self.code)
        
        code = formatCode(codeText)
        
        
        text = introText + '\n\n' + code + '\n\n'+ outroText
        
        if fileLocation is None:
            print(text)
            
        else:
            print('Saving code example to "{}".'.format(fileLocation))
            with open(fileLocation,'w') as f:
                f.write(text)

            print('Saving to local file')
            with open(os.path.join('Tutorials','tutorialList.txt'),'a') as f:
                f.write(fileLocation+'\n')
            



import sys,os

with open(os.path.join('docs','Tutorials','Keep.txt')) as f:
	keepFiles = [x.strip() for x in f.readlines()]

if not len(sys.argv)==1:
	folder = sys.argv[1]

	content = os.listdir(folder)
	if folder[-1]!=os.path.sep:
		folder+=os.path.sep
	if not len(content)==0:
		for f in content:
			if not f in keepFiles:
				try:
					os.remove(folder+f)
				except FileNotFoundError:
					pass
#			else:
#				print('Keeping: '+f)




